// High level management of views
// Copyright (C) 2004 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

static SELECTION_UNIVERSE *GetSelectionInView(EDITOR_VIEW *theView,SELECTION_UNIVERSE *theSelection)
// Work out a selection which is the part of the passed selection which
// intersects the visible view.
// If there is a problem, clean up, and return NULL
{
	UINT32
		topLine,
		numLines;
	INT32
		leftPixel;
	UINT32
		numPixels;
	CHUNK_HEADER
		*theChunk;
	UINT32
		theOffset,
		startPosition,
		endPosition;
	SELECTION_UNIVERSE
		*selection;

	if((selection=OpenSelectionUniverse()))						// get selection to work in
	{
		GetEditorViewTextInfo(theView,&topLine,&numLines,&leftPixel,&numPixels);

		LineToChunkPosition(theView->parentBuffer->textUniverse,topLine,&theChunk,&theOffset,&startPosition);		// get position of top line in this view
		LineToChunkPosition(theView->parentBuffer->textUniverse,topLine+numLines,&theChunk,&theOffset,&endPosition);		// get position past last line in this view

		SetSelectionRange(selection,startPosition,endPosition-startPosition);			// select the entire visible area of the view
		AndSelectionUniverse(theSelection,selection);			// get only the visible part of the current selection
	}
	return(selection);
}

void ViewEndSelectionChange(EDITOR_VIEW *theView)
// The selection in this view has been changed, update the view
{
	SELECTION_UNIVERSE
		*finalSelection;

	if(theView->heldSelection)									// make sure we did not fail to create this at the start
	{
		if((finalSelection=GetSelectionInView(theView,theView->parentBuffer->selectionUniverse)))	// get selection that intersects view
		{
			XorSelectionUniverse(theView->heldSelection,finalSelection);
			InvalidateViewSelection(theView,finalSelection);
			CloseSelectionUniverse(finalSelection);				// get rid of this
		}
		CloseSelectionUniverse(theView->heldSelection);			// get rid of memory this occupies
	}
	UpdateEditorWindows();

	ViewEndChange(theView);

	if(theView->viewSelectionChangedVector)						// see if someone needs to be called because of the change
	{
		theView->viewSelectionChangedVector(theView);			// call him
	}
}

void ViewStartSelectionChange(EDITOR_VIEW *theView)
// The selection in this view is about to change, so prepare for it
// This copies the visible part of the current selection, and compares
// it to the updated version when the end function is called
{
	theView->heldSelection=GetSelectionInView(theView,theView->parentBuffer->selectionUniverse);	// get a selection universe, use it to work out what changes are about to occur
	ViewStartChange(theView);
}

void ViewEndMarkChange(MARK_VISIBILITY *theVisibility)
// A visible mark in this view has been changed, update the view
{
	SELECTION_UNIVERSE
		*finalSelection;

	if(theVisibility->heldMark)									// make sure we did not fail to create this at the start
	{
		if((finalSelection=GetSelectionInView(theVisibility->parentView,theVisibility->parentMark->selectionUniverse)))	// work out where the mark is now
		{
			XorSelectionUniverse(theVisibility->heldMark,finalSelection);
			InvalidateViewSelection(theVisibility->parentView,finalSelection);
			CloseSelectionUniverse(finalSelection);				// get rid of this
		}
		CloseSelectionUniverse(theVisibility->heldMark);		// get rid of memory this occupies
	}
	UpdateEditorWindows();

	ViewEndChange(theVisibility->parentView);
}

void ViewStartMarkChange(MARK_VISIBILITY *theVisibility)
// A visible mark in this view is about to change, so prepare for it
{
	theVisibility->heldMark=GetSelectionInView(theVisibility->parentView,theVisibility->parentMark->selectionUniverse);	// get a selection universe, use it to work out what changes are about to occur
	ViewStartChange(theVisibility->parentView);
}

void ViewEndStyleAttributeChange(EDITOR_VIEW *theView)
// The attributes of styles in theView have finished changing
{
	ViewStyleFontsChanged(theView);
	InvalidateView(theView);									// this is pretty lazy
}

void ViewStartStyleAttributeChange(EDITOR_VIEW *theView)
// The attributes of styles in the view are being changed
{
}

void ViewEndStyleChange(EDITOR_VIEW *theView)
// The styles applied the text for this view have been changed, invalidate the view
{
	UINT32
		topLine,
		numLines;
	INT32
		leftPixel,
		xPosition;
	UINT32
		slopLeft,
		slopRight;
	UINT32
		numPixels;
	CHUNK_HEADER
		*theChunk;
	UINT32
		theOffset,
		currentPosition;
	UINT32
		lineLength,
		maxPosition;
	UINT32
		oldStyleStart,
		newStyleStart;
	UINT32
		oldStyleElements,
		newStyleElements;
	UINT32
		oldStyle,
		newStyle;
	bool
		hadOldStyle,
		hadNewStyle;
	UINT32
		i;
	bool
		done;

	if(theView->heldStyle)									// make sure we did not fail to create this at the start
	{
		GetEditorViewTextInfo(theView,&topLine,&numLines,&leftPixel,&numPixels);
		done=false;
		for(i=0;!done&&(i<numLines);i++)					// add invalidations a line at a time
		{
			LineToChunkPosition(theView->parentBuffer->textUniverse,topLine+i,&theChunk,&theOffset,&currentPosition);		// get position of top line in this view
			ChunkPositionToNextLine(theView->parentBuffer->textUniverse,theChunk,theOffset,&theChunk,&theOffset,&lineLength);	// get length of this line
			maxPosition=currentPosition+lineLength;			// go no further than this looking for a style change in this line

			while(!done&&(currentPosition<maxPosition))		// walk through the current line looking for a style change
			{
				hadOldStyle=GetStyleRange(theView->heldStyle,currentPosition,&oldStyleStart,&oldStyleElements,&oldStyle);
				hadNewStyle=GetStyleRange(theView->parentBuffer->styleUniverse,currentPosition,&newStyleStart,&newStyleElements,&newStyle);

				if(hadOldStyle||hadNewStyle)				// make sure there are some changes to look at
				{
					if(oldStyle!=newStyle)					// is there a change?
					{
						// currentPosition now contains the location of the style change, invalidate to the end of the line
						xPosition=leftPixel+numPixels;		// don't look past here
						GetEditorViewTextToGraphicPosition(theView,currentPosition,&xPosition,true,&slopLeft,&slopRight);
						if(xPosition>(INT32)slopLeft)
						{
							xPosition-=slopLeft;			// shift back to account for font overhang
						}
						else
						{
							xPosition=0;
						}
						if(xPosition<leftPixel)
						{
							xPosition=leftPixel;
						}
						xPosition-=leftPixel;				// make relative to scrolled view
						if(xPosition<(INT32)numPixels)		// work to do?
						{
							InvalidateViewPortion(theView,i,i+1,xPosition,numPixels);	// invalidate to the end of the view
						}
						currentPosition=maxPosition;		// done with this line
					}
					else
					{
						// move forward to the nearest change
						if(hadOldStyle)						// see if we had an old style
						{
							currentPosition=oldStyleStart+oldStyleElements;
							if(hadNewStyle)					// if had new style too, see if it changes sooner
							{
								if(currentPosition>newStyleStart+newStyleElements)
								{
									currentPosition=newStyleStart+newStyleElements;
								}
							}
						}
						else					// no old style, so must have had new
						{
							currentPosition=newStyleStart+newStyleElements;
						}
					}
				}
				else
				{
					done=true;								// no styles defined here, so no changes
				}
			}
		}
		CloseStyleUniverse(theView->heldStyle);				// get rid of memory this occupies
	}
	UpdateEditorWindows();
	ViewEndChange(theView);
}

void ViewStartStyleChange(EDITOR_VIEW *theView)
// The styles applied to text in this view are about to change, so prepare for it
{
	UINT32
		topLine,
		numLines;
	INT32
		leftPixel;
	UINT32
		numPixels;
	CHUNK_HEADER
		*theChunk;
	UINT32
		theOffset,
		startPosition,
		endPosition,
		actualStart,
		length;
	UINT32
		theStyle;
	UINT32
		i;
	bool
		done;

	if((theView->heldStyle=OpenStyleUniverse()))			// get a style universe, and add the styles visible in this view to it
	{
		GetEditorViewTextInfo(theView,&topLine,&numLines,&leftPixel,&numPixels);
		LineToChunkPosition(theView->parentBuffer->textUniverse,topLine,&theChunk,&theOffset,&startPosition);		// get position of top line in this view
		LineToChunkPosition(theView->parentBuffer->textUniverse,topLine+numLines,&theChunk,&theOffset,&endPosition);		// get position past last line in this view
		done=false;
		i=startPosition;
		while(i<endPosition&&!done)							// copy the part of the style array that intersects the view
		{
			if(GetStyleRange(theView->parentBuffer->styleUniverse,i,&actualStart,&length,&theStyle))
			{
				ForceStyleIntoRange(theStyle);
				if(actualStart<endPosition)
				{
					if(actualStart+length>endPosition)
					{
						length=endPosition-actualStart;
					}
					SetStyleRange(theView->heldStyle,actualStart,length,theStyle);
					i=actualStart+length;
				}
				else
				{
					done=true;
				}
			}
			else
			{
				done=true;
			}
		}
	}
}

void ViewEndTabSizeChange(EDITOR_VIEW *theView)
// The tab size on the passed view has finished changing
{
	InvalidateView(theView);								// this is pretty lazy
}

void ViewStartTabSizeChange(EDITOR_VIEW *theView)
// The tab size on the passed view is about to change
{
}

void ViewEndTextChange(EDITOR_VIEW *theView)
// The text in theView has finished being changed bring the view back
// up to date
{
	ViewEndChange(theView);
	if(theView->viewTextChangedVector)						// see if someone needs to be called because of the change
	{
		theView->viewTextChangedVector(theView);			// call him with my pointer
	}
}

void ViewStartTextChange(EDITOR_VIEW *theView)
// The text in theView is about to be changed in some way, do whatever is needed
// so that when ViewEndTextChange is called, the correct parts of the view are
// invalidated.
// NOTE: when the text changes, it is possible that the selection information
// will also change, so that updates will be handled correctly
{
	ViewStartChange(theView);
}

void ViewEndActiveChange(EDITOR_VIEW *theView)
// View's active status is done changing
{
	ViewEndSelectionChange(theView);
}

void ViewStartActiveChange(EDITOR_VIEW *theView)
// View's active status is changing
{
	ViewStartSelectionChange(theView);
}

static void UnSetViewStyleFont(EDITOR_VIEW *theView,UINT32 theStyle)
// Free the font currently in use by theStyle of theView
// NOTE: it is a bad idea to call this for the default style until the window
// will no longer be used, since the default is used when the given style
// is not available
{
	if(theView->viewStyles[theStyle].theFont)
	{
		FreeFont(theView->viewStyles[theStyle].theFont);
		theView->viewStyles[theStyle].theFont=NULL;				// mark this font as unused
	}
}

static bool SetViewStyleFont(EDITOR_VIEW *theView,UINT32 theStyle,char *fontName)
// Update the font used for a style in theView
// NOTE: it is necessary to call RecalculateViewFontInfo
// after calling this (that is left to the caller, since he may need to make
// many calls to this in a row and the recalculation can be left until after
// the last call)
{
	EDITOR_FONT
		*newFont;

	if((newFont=LoadFont(fontName)))
	{
		UnSetViewStyleFont(theView,theStyle);					// get rid of old font that was there (if any)
		theView->viewStyles[theStyle].theFont=newFont;
		return(true);
	}
	return(false);
}

static void SetViewTabSize(EDITOR_VIEW *theView,UINT32 theSize)
// set the tabSize for theView
{
	if(theSize>255)
	{
		theSize=255;			// limit this
	}
	theView->tabSize=theSize;
}

static void UnInitializeViewStyles(EDITOR_VIEW *theView)
// get rid of style information tied to theView
{
	UINT32
		i;

	for(i=0;i<VIEW_MAX_STYLES;i++)
	{
		UnSetViewStyleFont(theView,i);
	}
}

static bool InitializeViewStyles(EDITOR_VIEW *theView,EDITOR_VIEW_DESCRIPTOR *theDescriptor)
// initialize style information for theView
{
	UINT32
		i;

	for(i=0;i<VIEW_MAX_STYLES;i++)
	{
		theView->viewStyles[i].theFont=NULL;				// clear pointer
		theView->viewStyles[i].haveForegroundColor=false;	// no colors defined yet
		theView->viewStyles[i].haveBackgroundColor=false;
	}

	// set up default style (which must be present)
	theView->viewStyles[0].foregroundColor=theDescriptor->foregroundColor;
	theView->viewStyles[0].haveForegroundColor=true;

	theView->viewStyles[0].backgroundColor=theDescriptor->backgroundColor;
	theView->viewStyles[0].haveBackgroundColor=true;

	if(SetViewStyleFont(theView,0,theDescriptor->fontName))
	{
		return(true);
	}
	return(false);
}

UINT32 GetEditorViewTabSize(EDITOR_VIEW *theView)
// get the current tab size for the given view
{
	return(theView->tabSize);
}

bool SetEditorViewTabSize(EDITOR_VIEW *theView,UINT32 theSize)
// set the tab size to be used in theView
// if there is a problem, leave the tab size unchanged, set the error
// and return false
{
	ViewStartTabSizeChange(theView);
	SetViewTabSize(theView,theSize);
	ViewEndTabSizeChange(theView);
	return(true);
}

static void ForceStyleRange(UINT32 *theStyle)
// make sure the style is in range (for externally called routines)
{
	if(*theStyle>=VIEW_MAX_STYLES)
	{
		*theStyle=VIEW_MAX_STYLES-1;
	}
}

bool GetEditorViewStyleFont(EDITOR_VIEW *theView,UINT32 theStyle,char *theFont,UINT32 stringBytes)
// get the font that is in use in theView for the given style
// if the font will not fit within string bytes, return false
// if there is no font in use, return an empty string
{
	ForceStyleRange(&theStyle);
	if(stringBytes)
	{
		if(theView->viewStyles[theStyle].theFont)
		{
			GetFontName(theView->viewStyles[theStyle].theFont,theFont,stringBytes);
			return(true);
		}
		else
		{
			theFont[0]='\0';
			return(true);
		}
	}
	return(false);
}

bool SetEditorViewStyleFont(EDITOR_VIEW *theView,UINT32 theStyle,char *theFont)
// set the font to be used in theView
// this will force an update of theView
// if there is a problem, leave the window font unchanged, set the error
// and return false
{
	bool
		result;

	result=false;
	ForceStyleRange(&theStyle);
	ViewStartStyleAttributeChange(theView);
	if(SetViewStyleFont(theView,theStyle,theFont))				// set new font
	{
		result=true;
	}
	ViewEndStyleAttributeChange(theView);
	return(result);
}

void ClearEditorViewStyleFont(EDITOR_VIEW *theView,UINT32 theStyle)
// clear the font to be used for theStyle in theView
// this will force an update of theView
{
	ForceStyleRange(&theStyle);
	ViewStartStyleAttributeChange(theView);
	UnSetViewStyleFont(theView,theStyle);
	ViewEndStyleAttributeChange(theView);
}

bool GetEditorViewStyleForegroundColor(EDITOR_VIEW *theView,UINT32 theStyle,EDITOR_COLOR *foregroundColor)
// get the foreground color in use by theStyle of theView
// if the style does not have a foreground color associated with it, return false
{
	ForceStyleRange(&theStyle);
	if(theView->viewStyles[theStyle].haveForegroundColor)
	{
		*foregroundColor=theView->viewStyles[theStyle].foregroundColor;	// get pixel of foreground for this style
		return(true);
	}
	return(false);
}

bool SetEditorViewStyleForegroundColor(EDITOR_VIEW *theView,UINT32 theStyle,EDITOR_COLOR foregroundColor)
// set the foreground color to be used by theStyle of theView
// this will force an update of theView
// if there is a problem, leave the color unchanged, set the error
// and return false
{
	ForceStyleRange(&theStyle);
	ViewStartStyleAttributeChange(theView);
	theView->viewStyles[theStyle].foregroundColor=foregroundColor;
	theView->viewStyles[theStyle].haveForegroundColor=true;
	ViewEndStyleAttributeChange(theView);
	return(true);
}

void ClearEditorViewStyleForegroundColor(EDITOR_VIEW *theView,UINT32 theStyle)
// clear the foreground color to be used for theStyle in theView
// this will force an update of theView
{
	ForceStyleRange(&theStyle);
	ViewStartStyleAttributeChange(theView);
	theView->viewStyles[theStyle].haveForegroundColor=false;
	ViewEndStyleAttributeChange(theView);
}

bool GetEditorViewStyleBackgroundColor(EDITOR_VIEW *theView,UINT32 theStyle,EDITOR_COLOR *backgroundColor)
// get the background color in use by theStyle of theView
// if the style does not have a background color associated with it, return false
{
	ForceStyleRange(&theStyle);
	if(theView->viewStyles[theStyle].haveBackgroundColor)
	{
		*backgroundColor=theView->viewStyles[theStyle].backgroundColor;	// get pixel of background for this style
		return(true);
	}
	return(false);
}

bool SetEditorViewStyleBackgroundColor(EDITOR_VIEW *theView,UINT32 theStyle,EDITOR_COLOR backgroundColor)
// set the background color to be used by theStyle of theView
// this will force an update of theView
// if there is a problem, leave the color unchanged, set the error
// and return false
{
	ForceStyleRange(&theStyle);
	ViewStartStyleAttributeChange(theView);
	theView->viewStyles[theStyle].backgroundColor=backgroundColor;
	theView->viewStyles[theStyle].haveBackgroundColor=true;
	ViewEndStyleAttributeChange(theView);
	return(true);
}

void ClearEditorViewStyleBackgroundColor(EDITOR_VIEW *theView,UINT32 theStyle)
// clear the background color to be used for theStyle in theView
// this will force an update of theView
{
	ForceStyleRange(&theStyle);
	ViewStartStyleAttributeChange(theView);
	theView->viewStyles[theStyle].haveBackgroundColor=false;
	ViewEndStyleAttributeChange(theView);
}

bool GetEditorViewSelectionForegroundColor(EDITOR_VIEW *theView,EDITOR_COLOR *foregroundColor)
// get the foreground color in use by the selection of theView
// NOTE: it is possible that no color has been specified.
// If that is the case, this will return false
{
	if(theView->haveSelectionForegroundColor)
	{
		*foregroundColor=theView->selectionForegroundColor;
		return(true);
	}
	return(false);
}

bool SetEditorViewSelectionForegroundColor(EDITOR_VIEW *theView,EDITOR_COLOR foregroundColor)
// set the foreground color to be used for selections in theView
// this will force an update of theView
// if there is a problem, leave the color unchanged, set the error
// and return false
// NOTE: if haveForeground is false, this will clear the foreground color
// making the selection draw the foreground as the inverse of the current
// color
{
	theView->selectionForegroundColor=foregroundColor;
	theView->haveSelectionForegroundColor=true;
	InvalidateViewSelection(theView,theView->parentBuffer->selectionUniverse);
	return(true);
}

void ClearEditorViewSelectionForegroundColor(EDITOR_VIEW *theView)
// Clear the foreground color, making the selection draw the foreground
// as the inverse of the current color
{
	theView->haveSelectionForegroundColor=false;
	InvalidateViewSelection(theView,theView->parentBuffer->selectionUniverse);
}

bool GetEditorViewSelectionBackgroundColor(EDITOR_VIEW *theView,EDITOR_COLOR *backgroundColor)
// get the background color in use by the selection of theView
// NOTE: it is possible that no color has been specified.
// If that is the case, this will return false
{
	if(theView->haveSelectionBackgroundColor)
	{
		*backgroundColor=theView->selectionBackgroundColor;
		return(true);
	}
	return(false);
}

bool SetEditorViewSelectionBackgroundColor(EDITOR_VIEW *theView,EDITOR_COLOR backgroundColor)
// set the background color to be used for selections in theView
// this will force an update of theView
// if there is a problem, leave the color unchanged, set the error
// and return false
// NOTE: if haveBackground is false, this will clear the background color
// making the selection draw the background as the inverse of the current
// color
{
	theView->selectionBackgroundColor=backgroundColor;
	theView->haveSelectionBackgroundColor=true;
	InvalidateViewSelection(theView,theView->parentBuffer->selectionUniverse);
	return(true);
}

void ClearEditorViewSelectionBackgroundColor(EDITOR_VIEW *theView)
// Clear the background color, making the selection draw the background
// as the inverse of the current color
{
	theView->haveSelectionBackgroundColor=false;
	InvalidateViewSelection(theView,theView->parentBuffer->selectionUniverse);
}

void ViewsEndSelectionChange(EDITOR_BUFFER *theBuffer)
// The selection in theBuffer has finished being changed
// do whatever needs to be done
// to get it up to date
{
	EDITOR_VIEW
		*currentView;

	currentView=theBuffer->firstView;
	while(currentView)									// walk through all views, update each one as needed
	{
		ViewEndSelectionChange(currentView);
		currentView=currentView->nextBufferView;		// locate next view of this buffer
	}
}

void ViewsStartSelectionChange(EDITOR_BUFFER *theBuffer)
// The selection in theBuffer are about to be changed in some way, do whatever is needed
// NOTE: a change in the selection could mean something as simple as a cursor position change
// NOTE also: selection changes cannot be nested, and nothing except the selection is allowed
// to be altered during a selection change
{
	EDITOR_VIEW
		*currentView;

	currentView=theBuffer->firstView;
	while(currentView)									// walk through all views, update each one as needed
	{
		ViewStartSelectionChange(currentView);
		currentView=currentView->nextBufferView;		// locate next view of this buffer
	}
}

void ViewInvalidateMark(EDITOR_VIEW *theView,EDITOR_MARK *theMark)
// Invalidate the selection of the passed mark in the passed view
{
	InvalidateViewSelection(theView,theMark->selectionUniverse);
}

void ViewsInvalidateMark(EDITOR_MARK *theMark)
// Invalidate the selection of the passed mark in all visible views
{
	MARK_VISIBILITY
		*currentVisibility;

	currentVisibility=theMark->headVisibility;
	while(currentVisibility)							// walk through all views, update each one as needed
	{
		ViewInvalidateMark(currentVisibility->parentView,theMark);
		currentVisibility=currentVisibility->nextView;	// locate next view this mark is visible in
	}
}

void ViewsEndMarkChange(EDITOR_MARK *theMark)
// The passed mark has finished being changed, update all
// affected views.
{
	MARK_VISIBILITY
		*currentVisibility;

	currentVisibility=theMark->headVisibility;
	while(currentVisibility)							// walk through all views, update each one as needed
	{
		ViewEndMarkChange(currentVisibility);
		currentVisibility=currentVisibility->nextView;	// locate next view this mark is visible in
	}
}

void ViewsStartMarkChange(EDITOR_MARK *theMark)
// Marks (which may be visible) in theBuffer are about to be changed in some way, do whatever is needed
// NOTE: mark changes cannot be nested, and nothing except a mark is allowed
// to be altered during a mark change
{
	MARK_VISIBILITY
		*currentVisibility;

	currentVisibility=theMark->headVisibility;
	while(currentVisibility)							// walk through all views, update each one as needed
	{
		ViewStartMarkChange(currentVisibility);
		currentVisibility=currentVisibility->nextView;	// locate next view this mark is visible in
	}
}

void ViewsStartStyleChange(EDITOR_BUFFER *theBuffer)
// The style in theBuffer is about to be changed in some way, do whatever is needed
// NOTE also: style changes cannot be nested, and nothing except the style is allowed
// to be altered during a style change
{
	EDITOR_VIEW
		*currentView;

	currentView=theBuffer->firstView;
	while(currentView)									// walk through all views, update each one as needed
	{
		ViewStartStyleChange(currentView);
		currentView=currentView->nextBufferView;		// locate next view of this buffer
	}
}

void ViewsEndStyleChange(EDITOR_BUFFER *theBuffer)
// The style in theBuffer has finished being changed
// do whatever needs to be done to get it up to date
{
	EDITOR_VIEW
		*currentView;

	currentView=theBuffer->firstView;
	while(currentView)									// walk through all views, update each one as needed
	{
		ViewEndStyleChange(currentView);
		currentView=currentView->nextBufferView;		// locate next view of this buffer
	}
}

void ViewsStartTextChange(EDITOR_BUFFER *theBuffer)
// The text in theBuffer is about to be changed in some way, do whatever is needed
// NOTE: when the text changes, it is possible that the selection information
// will also change, so that updates will be handled correctly
{
	EDITOR_VIEW
		*currentView;

	currentView=theBuffer->firstView;
	while(currentView)									// walk through all views, update each one as needed
	{
		ViewStartTextChange(currentView);
		currentView=currentView->nextBufferView;		// locate next view of this buffer
	}
}

void ViewsEndTextChange(EDITOR_BUFFER *theBuffer)
// The text in theBuffer has finished being changed, and the views have been
// updated, do whatever needs to be done
{
	EDITOR_VIEW
		*currentView;

	currentView=theBuffer->firstView;
	while(currentView)								// walk through all views, update each one as needed
	{
		ViewEndTextChange(currentView);
		currentView=currentView->nextBufferView;		// locate next view of this buffer
	}
}

void EditorActivateView(EDITOR_VIEW *theView)
// activate theView (this is not nestable)
{
	ViewStartActiveChange(theView);
	theView->active=true;
	ViewEndActiveChange(theView);
}

void EditorDeactivateView(EDITOR_VIEW *theView)
// deactivate theView (this is not nestable)
{
	ViewStartActiveChange(theView);
	theView->active=false;
	ViewEndActiveChange(theView);
}

void DisposeEditorView(EDITOR_VIEW *theView)
// unlink theView from everywhere, and dispose of it
{
	DisposeViewInfo(theView);							// tell GUI to tear it down
	UnInitializeViewStyles(theView);					// get rid of style information for theView
	UnlinkVisibleMarksFromView(theView);				// get rid of any visible marks hanging off the view
	UnlinkViewFromBuffer(theView);
	MDisposePtr(theView);
}

EDITOR_VIEW *CreateEditorView(EDITOR_WINDOW *theWindow,EDITOR_VIEW_DESCRIPTOR *theDescriptor)
// create a view in theWindow using theDescriptor
// if there is a problem, SetError, and return NULL
{
	EDITOR_VIEW
		*theView;

	if((theView=(EDITOR_VIEW *)MNewPtr(sizeof(EDITOR_VIEW))))
	{
		LinkViewToBuffer(theView,theDescriptor->theBuffer);
		theView->parentWindow=theWindow;
		theView->active=theDescriptor->active;
		theView->wantHome=false;
		SetViewTabSize(theView,theDescriptor->tabSize);	// set, and limit the tab size
		theView->haveSelectionForegroundColor=false;	// select by inverting text colors by default
		theView->haveSelectionBackgroundColor=false;
		theView->headVisibility=NULL;					// no visible marks hanging off of this view yet
		theView->tailVisibility=NULL;
		theView->viewTextChangedVector=theDescriptor->viewTextChangedVector;
		theView->viewSelectionChangedVector=theDescriptor->viewSelectionChangedVector;
		theView->viewPositionChangedVector=theDescriptor->viewPositionChangedVector;

		if(InitializeViewStyles(theView,theDescriptor))
		{
			if(CreateViewInfo(theView,theDescriptor))
			{
				return(theView);
			}
		}
		UnlinkViewFromBuffer(theView);
		MDisposePtr(theView);
	}
	return(NULL);
}
