// Document window handling
// Copyright (C) 2000 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

typedef struct
{
	EDITOR_VIEW
		*theView;
	bool
		verticalScrollBarOnLeft;				// gives the preference for vertical scroll bar placement
	SCROLL_BAR
		*verticalScrollBar,
		*horizontalScrollBar;
	EDITOR_FONT
		*statusBarFont;							// info about currently selected status bar font
	EDITOR_COLOR
		backgroundColor,						// color to use when drawing background
		foregroundColor;						// color to use when drawing foreground
} GUI_DOCUMENT_WINDOW_INFO;

static void SetDocumentForegroundColor(EDITOR_WINDOW *theWindow,EDITOR_COLOR foreground)
// set the foreground color used to draw theWindow
{
	GUI_DOCUMENT_WINDOW_INFO
		*theWindowInfo;

	theWindowInfo=(GUI_DOCUMENT_WINDOW_INFO *)theWindow->windowInfo;
	theWindowInfo->foregroundColor=foreground;
}

static void SetDocumentBackgroundColor(EDITOR_WINDOW *theWindow,EDITOR_COLOR background)
// set the background color used to draw theWindow
{
	GUI_DOCUMENT_WINDOW_INFO
		*theWindowInfo;

	theWindowInfo=(GUI_DOCUMENT_WINDOW_INFO *)theWindow->windowInfo;
	theWindowInfo->backgroundColor=background;
}

static void GetDocumentWindowRects(EDITOR_WINDOW *theWindow,EDITOR_RECT *contentRect,EDITOR_RECT *viewRect,EDITOR_RECT *verticalRect,EDITOR_RECT *horizontalRect,EDITOR_RECT *statusRect)
// get the rectangles for all the areas of a document window
{
	XWindowAttributes
		theAttributes;
	EDITOR_FONT
		*statusBarFont;							// info about currently selected status bar font
	UINT32
		statusHeight;
	GUI_DOCUMENT_WINDOW_INFO
		*theWindowInfo;

	theWindowInfo=(GUI_DOCUMENT_WINDOW_INFO *)theWindow->windowInfo;
	XGetWindowAttributes(xDisplay,((WINDOW_LIST_ELEMENT *)theWindow->userData)->xWindow,&theAttributes);

	contentRect->x=theAttributes.border_width;
	contentRect->y=theAttributes.border_width;
	contentRect->w=theAttributes.width;
	contentRect->h=theAttributes.height;

	statusRect->x=theAttributes.border_width;
	statusRect->y=theAttributes.border_width;
	statusRect->w=theAttributes.width;

	statusBarFont=((GUI_DOCUMENT_WINDOW_INFO *)theWindow->windowInfo)->statusBarFont;
	statusHeight=statusBarFont->ascent+statusBarFont->descent;

	statusRect->h=statusHeight+STATUS_BAR_LINE_HEIGHT;

	viewRect->x=theAttributes.border_width;					// move over for the left border
	viewRect->y=statusRect->y+statusRect->h;				// leave room for status bar

	if(theWindowInfo->verticalScrollBarOnLeft)
	{
		viewRect->x+=SCROLL_BAR_WIDTH;						// if scroll bar on left, then step over it to begin the view
	}

	if(theAttributes.width>SCROLL_BAR_WIDTH)
	{
		viewRect->w=theAttributes.width-SCROLL_BAR_WIDTH;		// pull back for scroll bar, and right free area
	}
	else
	{
		viewRect->w=0;
	}
	if((UINT32)theAttributes.height>(statusHeight+STATUS_BAR_LINE_HEIGHT+SCROLL_BAR_WIDTH))
	{
		viewRect->h=theAttributes.height-statusHeight-STATUS_BAR_LINE_HEIGHT-SCROLL_BAR_WIDTH;		// pull back for bottom scroll bar, and bottom free area
	}
	else
	{
		viewRect->h=0;										// view cant fit!
	}

	if(theWindowInfo->verticalScrollBarOnLeft)
	{
		verticalRect->x=theAttributes.border_width;			// if on left, then place it just after the border
	}
	else
	{
		verticalRect->x=viewRect->x+viewRect->w;			// if on right, place just after the view
	}

	verticalRect->y=viewRect->y;
	verticalRect->w=SCROLL_BAR_WIDTH;
	verticalRect->h=viewRect->h;

	horizontalRect->x=viewRect->x;
	horizontalRect->y=viewRect->y+viewRect->h;
	horizontalRect->w=viewRect->w;
	horizontalRect->h=SCROLL_BAR_WIDTH;
}

static void CreateStatusString(EDITOR_WINDOW *theWindow,char *theString)
// create status string for theWindow
// NOTE: theString must be large enough to hold the status
{
	EDITOR_BUFFER
		*theBuffer;
	UINT32
		startPosition,						// selection information
		endPosition,
		startLine,
		endLine,
		startLinePosition,
		endLinePosition,
		totalSegments,
		totalSpan;

	theBuffer=theWindow->theBuffer;

	EditorGetSelectionInfo(theBuffer,theBuffer->selectionUniverse,&startPosition,&endPosition,&startLine,&endLine,&startLinePosition,&endLinePosition,&totalSegments,&totalSpan);

	sprintf(theString,"%s%sLines:%d Chars:%d",
			AtUndoCleanPoint(theBuffer)?"  ":"* ",
			theWindow->theBuffer->theTask?"<Task running> ":"",
			(int)theBuffer->textUniverse->totalLines,
			(int)theBuffer->textUniverse->totalBytes
			);

	if(!totalSegments)
	{
		sprintf(&theString[strlen(theString)]," Cursor(%d:%d %d)",(int)startLine+1,(int)startLinePosition,(int)startPosition);
	}
	else
	{
		sprintf(&theString[strlen(theString)]," Selections: Number:%d TotalChars:%d Start(%d:%d %d) End(%d:%d %d)",
				(int)totalSegments,
				(int)totalSpan,
				(int)startLine+1,(int)startLinePosition,(int)startPosition,
				(int)endLine+1,(int)endLinePosition,(int)endPosition
				);
	}
}

static void DrawStatusBar(EDITOR_WINDOW *theWindow)
// draw the status bar for theWindow
{
	int
		linePosition;
	XRectangle
		regionBox;
	WINDOW_LIST_ELEMENT
		*theWindowElement;
	GC
		graphicsContext;
	Region
		invalidRegion;
	EDITOR_RECT
		contentRect,
		viewRect,
		verticalRect,
		horizontalRect,
		statusRect;
	EDITOR_FONT
		*statusBarFont;
	int
		stringWidth;
	char
		statusString[256];											// the status string

	theWindowElement=(WINDOW_LIST_ELEMENT *)theWindow->userData;		// get the x window information associated with this document window
	graphicsContext=theWindowElement->graphicsContext;				// get graphics context for this window

	invalidRegion=XCreateRegion();

	GetDocumentWindowRects(theWindow,&contentRect,&viewRect,&verticalRect,&horizontalRect,&statusRect);

	regionBox.x=statusRect.x;
	regionBox.y=statusRect.y;
	regionBox.width=statusRect.w;
	regionBox.height=statusRect.h;
	XUnionRectWithRegion(&regionBox,invalidRegion,invalidRegion);		// create a region which is the rectangle of the entire status bar
	XIntersectRegion(invalidRegion,theWindowElement->invalidRegion,invalidRegion);	// intersect with what is invalid in this window

	if(!XEmptyRegion(invalidRegion))								// see if it has an invalid area
	{
		XSetRegion(xDisplay,graphicsContext,invalidRegion);			// set clipping to what's invalid
		XSetFillStyle(xDisplay,graphicsContext,FillSolid);
		XClipBox(invalidRegion,&regionBox);							// get bounds of invalid region

		CreateStatusString(theWindow,&statusString[0]);
		statusBarFont=((GUI_DOCUMENT_WINDOW_INFO *)theWindow->windowInfo)->statusBarFont;
		XSetFont(xDisplay,graphicsContext,statusBarFont->theXFont->fid);		// point to the font

		stringWidth=XTextWidth(statusBarFont->theXFont,&statusString[0],strlen(statusString));

		XSetForeground(xDisplay,graphicsContext,EditorColorToXPixel(statusBackgroundColor));
		XFillRectangle(xDisplay,theWindowElement->xWindow,graphicsContext,regionBox.x,regionBox.y,regionBox.width,regionBox.height);	// fill with white
		XSetForeground(xDisplay,graphicsContext,EditorColorToXPixel(statusForegroundColor));
		XDrawString(xDisplay,theWindowElement->xWindow,graphicsContext,statusRect.x+VIEW_LEFT_MARGIN,statusRect.y+statusBarFont->ascent,&statusString[0],strlen(statusString));

		linePosition=statusRect.y+statusRect.h-STATUS_BAR_LINE_HEIGHT;
		XDrawLine(xDisplay,theWindowElement->xWindow,graphicsContext,statusRect.x,linePosition,statusRect.w,linePosition);
		linePosition+=STATUS_BAR_LINE_HEIGHT-1;
		XSetForeground(xDisplay,graphicsContext,EditorColorToXPixel(WHITE));
		XDrawLine(xDisplay,theWindowElement->xWindow,graphicsContext,statusRect.x,linePosition,statusRect.w,linePosition);
		XSetClipMask(xDisplay,graphicsContext,None);				// get rid of clip mask
	}
	XDestroyRegion(invalidRegion);									// get rid of invalid region for the view
}

static void DrawGrowArea(EDITOR_WINDOW *theWindow)
// draw the grow box area of the window
{
	WINDOW_LIST_ELEMENT
		*theWindowElement;
	GC
		graphicsContext;
	XRectangle
		regionBox;
	EDITOR_RECT
		contentRect,
		viewRect,
		verticalRect,
		horizontalRect,
		statusRect;

	theWindowElement=(WINDOW_LIST_ELEMENT *)theWindow->userData;			// get the x window information associated with this document window
	graphicsContext=theWindowElement->graphicsContext;					// get graphics context for this window

	GetDocumentWindowRects(theWindow,&contentRect,&viewRect,&verticalRect,&horizontalRect,&statusRect);

	regionBox.x=verticalRect.x;
	regionBox.y=horizontalRect.y;
	regionBox.width=verticalRect.w;
	regionBox.height=horizontalRect.h;

	XSetFillStyle(xDisplay,graphicsContext,FillSolid);
	XSetForeground(xDisplay,graphicsContext,EditorColorToXPixel(0x868693));
	XFillRectangle(xDisplay,theWindowElement->xWindow,graphicsContext,regionBox.x,regionBox.y,regionBox.width,regionBox.height);	// fill with white
}

void AdjustDocumentWindowStatus(EDITOR_WINDOW *theWindow)
// adjust the status of the document window
// this includes updating the status bar, icon, and anything else status related
{
	EDITOR_RECT
		contentRect,
		viewRect,
		verticalRect,
		horizontalRect,
		statusRect;
	XWMHints
		theHints;

	GetDocumentWindowRects(theWindow,&contentRect,&viewRect,&verticalRect,&horizontalRect,&statusRect);
	statusRect.h-=STATUS_BAR_LINE_HEIGHT;								// just invalidate the status area
	InvalidateWindowRect(theWindow,&statusRect);

	// adjust the pixmap for the icon based on the window's "clean" status
	theHints.flags=IconPixmapHint;
	if(AtUndoCleanPoint(theWindow->theBuffer))
	{
		theHints.icon_pixmap=documentIconPixmap;				// point at document icon
	}
	else
	{
		theHints.icon_pixmap=documentModifiedIconPixmap;
	}
	if(theHints.icon_pixmap!=((WINDOW_LIST_ELEMENT *)theWindow->userData)->currentIcon)
	{
		((WINDOW_LIST_ELEMENT *)theWindow->userData)->currentIcon=theHints.icon_pixmap;	// remember this so we can tell if it changed
		XSetWMHints(xDisplay,((WINDOW_LIST_ELEMENT *)theWindow->userData)->xWindow,&theHints);
	}
}

static void AdjustDocumentWindowViewStatus(EDITOR_VIEW *theView)
// adjust the status bar of the document window with the given view
{
	AdjustDocumentWindowStatus(theView->parentWindow);
}

static void AdjustDocumentWindowScrollThumbs(EDITOR_VIEW *theView)
// look at the data of the passed view, and
// adjust the thumbs of its scroll bars so that
// they are correct
{
	UINT32
		scrollBarLeftPixel;

	AdjustScrollBar(((GUI_DOCUMENT_WINDOW_INFO *)theView->parentWindow->windowInfo)->verticalScrollBar,theView->parentBuffer->textUniverse->totalLines,((GUI_VIEW_INFO *)theView->viewInfo)->topLine);
	// NOTE: since views allow for having a left pixel <0, but the scroll bars do not, peg the bar at 0
	if(((GUI_VIEW_INFO *)theView->viewInfo)->leftPixel>=0)
	{
		scrollBarLeftPixel=(UINT32)((GUI_VIEW_INFO *)theView->viewInfo)->leftPixel;
	}
	else
	{
		scrollBarLeftPixel=0;
	}
	// since the view is allowed to scroll beyond the length of the horizontal scroll bar, pin the bar at the limit
	if(((GUI_VIEW_INFO *)theView->viewInfo)->leftPixel>H_SCROLL_MAX_WIDTH)
	{
		scrollBarLeftPixel=H_SCROLL_MAX_WIDTH;
	}
	AdjustScrollBar(((GUI_DOCUMENT_WINDOW_INFO *)theView->parentWindow->windowInfo)->horizontalScrollBar,H_SCROLL_MAX_WIDTH,scrollBarLeftPixel);
}

static void AdjustDocumentWindowScrollThumbsAndStatus(EDITOR_VIEW *theView)
// adjust the parts of the window that change when the view changes
{
	AdjustDocumentWindowScrollThumbs(theView);
	AdjustDocumentWindowViewStatus(theView);
}

static void DrawDocumentWindow(EDITOR_WINDOW *theWindow)
// a document window needs to be updated, this
// will draw it
// NOTE: the window's invalidRegion is set to the area which is in need of update
{
	DrawStatusBar(theWindow);
	DrawScrollBar(((GUI_DOCUMENT_WINDOW_INFO *)theWindow->windowInfo)->verticalScrollBar);		// draw the scroll bars
	DrawScrollBar(((GUI_DOCUMENT_WINDOW_INFO *)theWindow->windowInfo)->horizontalScrollBar);
	DrawGrowArea(theWindow);
	DrawView(((GUI_DOCUMENT_WINDOW_INFO *)theWindow->windowInfo)->theView);	// draw this view
}

void UpdateDocumentWindow(EDITOR_WINDOW *theWindow)
// make sure the window is up to date
{
	WINDOW_LIST_ELEMENT
		*theWindowElement;

	theWindowElement=(WINDOW_LIST_ELEMENT *)theWindow->userData;				// get the x window information associated with this document window
	if(!XEmptyRegion(theWindowElement->invalidRegion))						// see if it has an invalid area
	{
		DrawDocumentWindow(theWindow);
		XDestroyRegion(theWindowElement->invalidRegion);					// clear out the invalid region now that the window has been drawn
		theWindowElement->invalidRegion=XCreateRegion();
	}
}

void AdjustDocumentWindowForNewSize(EDITOR_WINDOW *theWindow)
// if the window has changed size, then the
// controls will be needing update, so adjust them for the new
// size
{
	EDITOR_RECT
		contentRect,
		viewRect,
		verticalRect,
		horizontalRect,
		statusRect;

	GetDocumentWindowRects(theWindow,&contentRect,&viewRect,&verticalRect,&horizontalRect,&statusRect);

	SetViewBounds(((GUI_DOCUMENT_WINDOW_INFO *)theWindow->windowInfo)->theView,&viewRect);		// change the bounds of the view

	RepositionScrollBar(((GUI_DOCUMENT_WINDOW_INFO *)theWindow->windowInfo)->verticalScrollBar,verticalRect.x,verticalRect.y,verticalRect.h,false);
	RepositionScrollBar(((GUI_DOCUMENT_WINDOW_INFO *)theWindow->windowInfo)->horizontalScrollBar,horizontalRect.x,horizontalRect.y,horizontalRect.w,true);
}

void DocumentWindowButtonPress(EDITOR_WINDOW *theWindow,XEvent *theEvent)
// handle button press events for document windows (see what the user is clicking
// on, and attempt to deal with it)
{
	WINDOW_LIST_ELEMENT
		*theWindowElement;

	theWindowElement=(WINDOW_LIST_ELEMENT *)theWindow->userData;
	if(!TrackView(((GUI_DOCUMENT_WINDOW_INFO *)theWindow->windowInfo)->theView,theEvent))
	{
		if(!TrackScrollBar(((GUI_DOCUMENT_WINDOW_INFO *)theWindow->windowInfo)->verticalScrollBar,theEvent))
		{
			TrackScrollBar(((GUI_DOCUMENT_WINDOW_INFO *)theWindow->windowInfo)->horizontalScrollBar,theEvent);
		}
	}
}

void DocumentPeriodicProc(EDITOR_WINDOW *theWindow)
// this is called roughly twice a second to flash cursors, or do whatever in the document
// when the document has focus
{
	ViewCursorTask(((GUI_DOCUMENT_WINDOW_INFO *)theWindow->windowInfo)->theView);			// flash the cursor in the view
}

void DocumentWindowEvent(EDITOR_WINDOW *theWindow,XEvent *theEvent)
// handle an event arriving for a document window
{
	WINDOW_LIST_ELEMENT
		*theWindowElement;

	theWindowElement=(WINDOW_LIST_ELEMENT *)theWindow->userData;
	switch(theEvent->type)
	{
		case ButtonPress:
			XSetInputFocus(xDisplay,theWindowElement->xWindow,RevertToPointerRoot,theEvent->xbutton.time);	// give it the focus
			DocumentWindowButtonPress(theWindow,theEvent);										// go handle the button press
			break;
		case KeyPress:
			HandleViewKeyEvent(((GUI_DOCUMENT_WINDOW_INFO *)theWindowElement->theEditorWindow->windowInfo)->theView,theEvent);
			break;
		case ClientMessage:
			if(theEvent->xclient.send_event)
			{
				if(theEvent->xclient.data.l[0]==(int)deleteWindowAtom)
				{
					HandleShellCommand("close",1,&(theWindow->theBuffer->contentName));		// tell shell to close this
				}
			}
			break;
		default:
			break;
	}
}

void ScrollEditViewEvent(EDITOR_VIEW *theView,UINT32 eventCode,bool repeating)
// send a scroll event to the view
{
	VIEW_EVENT
		theRecord;
	EDITOR_KEY
		theKeyData;

	if(!repeating)
	{
		theRecord.eventType=VET_KEYDOWN;
	}
	else
	{
		theRecord.eventType=VET_KEYREPEAT;
	}
	theRecord.theView=theView;
	theKeyData.isVirtual=true;
	theKeyData.keyCode=eventCode;		// this is the key
	theKeyData.modifiers=0;				// no modifiers
	theRecord.eventData=&theKeyData;
	HandleViewEvent(&theRecord);
}

static void ScrollEditViewUpEvent(SCROLL_BAR *theScrollBar,bool repeating,void *parameters)
// tell the editor that the view in the window with theScrollBar is to be scrolled up
{
	ScrollEditViewEvent((EDITOR_VIEW *)parameters,VVK_SCROLLUP,repeating);
}

static void ScrollEditViewDownEvent(SCROLL_BAR *theScrollBar,bool repeating,void *parameters)
// tell the editor that the view in the window with theScrollBar is to be scrolled down
{
	ScrollEditViewEvent((EDITOR_VIEW *)parameters,VVK_SCROLLDOWN,repeating);
}

static void ScrollEditViewLeftEvent(SCROLL_BAR *theScrollBar,bool repeating,void *parameters)
// tell the editor that the view in the window with theScrollBar is to be scrolled left
{
	ScrollEditViewEvent((EDITOR_VIEW *)parameters,VVK_SCROLLLEFT,repeating);
}

static void ScrollEditViewRightEvent(SCROLL_BAR *theScrollBar,bool repeating,void *parameters)
// tell the editor that the view in the window with theScrollBar is to be scrolled right
{
	ScrollEditViewEvent((EDITOR_VIEW *)parameters,VVK_SCROLLRIGHT,repeating);
}

static void ScrollEditViewPageUpEvent(SCROLL_BAR *theScrollBar,bool repeating,void *parameters)
// tell the editor that the view in the window with theScrollBar is to be paged up
{
	ScrollEditViewEvent((EDITOR_VIEW *)parameters,VVK_DOCUMENTPAGEUP,repeating);
}

static void ScrollEditViewPageDownEvent(SCROLL_BAR *theScrollBar,bool repeating,void *parameters)
// tell the editor that the view in the window with theScrollBar is to be paged down
{
	ScrollEditViewEvent((EDITOR_VIEW *)parameters,VVK_DOCUMENTPAGEDOWN,repeating);
}

static void ScrollEditViewPageLeftEvent(SCROLL_BAR *theScrollBar,bool repeating,void *parameters)
// tell the editor that the view in the window with theScrollBar is to be paged left
{
	ScrollEditViewEvent((EDITOR_VIEW *)parameters,VVK_DOCUMENTPAGELEFT,repeating);
}

static void ScrollEditViewPageRightEvent(SCROLL_BAR *theScrollBar,bool repeating,void *parameters)
// tell the editor that the view in the window with theScrollBar is to be paged right
{
	ScrollEditViewEvent((EDITOR_VIEW *)parameters,VVK_DOCUMENTPAGERIGHT,repeating);
}

static void ScrollEditViewThumbVerticalEvent(SCROLL_BAR *theScrollBar,UINT32 newValue,void *parameters)
// tell the editor that the view in the window with theScrollBar is to be positioned vertically
{
	VIEW_EVENT
		theRecord;
	VIEW_POS_EVENT_DATA
		thePosData;

	theRecord.eventType=VET_POSITIONVERTICAL;
	theRecord.theView=(EDITOR_VIEW *)parameters;
	thePosData.modifiers=0;
	thePosData.position=newValue;
	theRecord.eventData=&thePosData;
	HandleViewEvent(&theRecord);
}

static void ScrollEditViewThumbHorizontalEvent(SCROLL_BAR *theScrollBar,UINT32 newValue,void *parameters)
// tell the editor that the view in the window with theScrollBar is to be positioned horizontally
{
	VIEW_EVENT
		theRecord;
	VIEW_POS_EVENT_DATA
		thePosData;

	theRecord.eventType=VET_POSITIONHORIZONTAL;
	theRecord.theView=(EDITOR_VIEW *)parameters;
	thePosData.modifiers=0;
	thePosData.position=newValue;
	theRecord.eventData=&thePosData;
	HandleViewEvent(&theRecord);
}

static void DisposeDocumentWindow(EDITOR_WINDOW *theWindow)
// close a given window, kill off the view
{
	WINDOW_LIST_ELEMENT
		*theWindowElement;

	theWindowElement=(WINDOW_LIST_ELEMENT *)theWindow->userData;		// get the x window associated with this document window
	DisposeScrollBar(((GUI_DOCUMENT_WINDOW_INFO *)theWindow->windowInfo)->horizontalScrollBar);
	DisposeScrollBar(((GUI_DOCUMENT_WINDOW_INFO *)theWindow->windowInfo)->verticalScrollBar);
	DisposeEditorView(((GUI_DOCUMENT_WINDOW_INFO *)theWindow->windowInfo)->theView);	// get rid of the view, unlink it from everywhere
	XDestroyRegion(theWindowElement->invalidRegion);				// destroy invalid region
	XFreeGC(xDisplay,theWindowElement->graphicsContext);			// get rid of the graphics context for this window
	FocusAwayFrom(theWindow);										// move focus off this window
	XDestroyWindow(xDisplay,theWindowElement->xWindow);				// tell x to make the window go away
	WaitUntilDestroyed(theWindow);
	UnlinkWindowElement(theWindowElement);							// unlink it from our list
	MDisposePtr(theWindowElement);
	FreeFont(((GUI_DOCUMENT_WINDOW_INFO *)theWindow->windowInfo)->statusBarFont);
	MDisposePtr(theWindow->windowInfo);
	MDisposePtr(theWindow);
}

static EDITOR_WINDOW *CreateDocumentWindow(EDITOR_BUFFER *theBuffer,EDITOR_RECT *theRect,char *theTitle,char *fontName,UINT32 tabSize,EDITOR_COLOR foreground,EDITOR_COLOR background)
// open a new document window with one EDITOR_VIEW onto theBuffer
// set the view(s) to use the given defaults
{
	EDITOR_RECT
		contentRect,
		viewRect,
		verticalRect,
		horizontalRect,
		statusRect;
	EDITOR_WINDOW
		*theWindow;
	GUI_DOCUMENT_WINDOW_INFO
		*theDocumentWindowInfo;
	WINDOW_LIST_ELEMENT
		*theWindowElement;
	SCROLL_BAR_DESCRIPTOR
		scrollDescriptor;
	EDITOR_VIEW_DESCRIPTOR
		viewDescriptor;
	XSetWindowAttributes
		theAttributes;
	UINT32
		valueMask;
	Atom
		protocols[2];																	// number of protocols
	XWMHints
		theHints;																		// hints to tell window manager how to treat these windows
	XSizeHints
		theSizeHints;																	// window size info (used to REALLY convince X that I know where I want my own window)

	if((theWindow=(EDITOR_WINDOW *)MNewPtr(sizeof(EDITOR_WINDOW))))						// create editor window
	{
		theWindow->theBuffer=theBuffer;

		if((theDocumentWindowInfo=(GUI_DOCUMENT_WINDOW_INFO *)MNewPtr(sizeof(GUI_DOCUMENT_WINDOW_INFO))))	// make structure to contain document window info
		{
			if((theDocumentWindowInfo->statusBarFont=LoadFont(defaultStatusBarFont)))			// load in the font
			{
				if((theWindowElement=(WINDOW_LIST_ELEMENT *)MNewPtr(sizeof(WINDOW_LIST_ELEMENT))))	// create a window list element
				{
					LinkWindowElement(theWindowElement);								// link it to the global window list that we maintain
					theWindowElement->theEditorWindow=theWindow;						// point to our high level concept of window
					theWindow->windowType=EWT_DOCUMENT;									// set the editor window type to "document"
					theWindow->userData=theWindowElement;								// point back to the window element
					theWindow->windowInfo=theDocumentWindowInfo;						// point to the edit info

					SetDocumentForegroundColor(theWindow,foreground);					// set the colors for this window
					SetDocumentBackgroundColor(theWindow,background);

					theAttributes.background_pixmap=None;
					theAttributes.background_pixel=EditorColorToXPixel(theDocumentWindowInfo->backgroundColor);
					theAttributes.border_pixmap=None;
					theAttributes.border_pixel=EditorColorToXPixel(theDocumentWindowInfo->foregroundColor);
					valueMask=CWBackPixmap|CWBackPixel|CWBorderPixmap|CWBorderPixel;

					// Due to X's nasty error handling, we assume this create does not fail!!!

					theWindowElement->realTopWindow=theWindowElement->xWindow=XCreateWindow(xDisplay,RootWindow(xDisplay,xScreenNum),theRect->x,theRect->y,theRect->w,theRect->h,XWINDOW_BORDER_WIDTH,CopyFromParent,InputOutput,DefaultVisual(xDisplay,xScreenNum),valueMask,&theAttributes);

					theSizeHints.flags=USSize|USPosition|PMinSize;
					theSizeHints.x=theRect->x;
					theSizeHints.y=theRect->y;
					theSizeHints.width=theRect->w;
					theSizeHints.height=theRect->h;
					theSizeHints.min_width=1;
					theSizeHints.min_height=1;
					XSetWMNormalHints(xDisplay,theWindowElement->xWindow,&theSizeHints);

					XStoreName(xDisplay,theWindowElement->xWindow,theTitle);
					XSetIconName(xDisplay,theWindowElement->xWindow,theTitle);
					XSelectInput(xDisplay,theWindowElement->xWindow,ExposureMask|ButtonPressMask|ButtonReleaseMask|FocusChangeMask|KeyPressMask|StructureNotifyMask);	// want to see these events

					protocols[0]=takeFocusAtom;
					protocols[1]=deleteWindowAtom;
					XSetWMProtocols(xDisplay,theWindowElement->xWindow,&(protocols[0]),2);	// tell X we would like to participate in these protocols

					theHints.flags=InputHint|StateHint|WindowGroupHint;
					theHints.input=True;
					theHints.initial_state=NormalState;
					theHints.window_group=((WINDOW_LIST_ELEMENT *)rootMenuWindow->userData)->xWindow;	// all documents belong to the group headed by the main menu window
					XSetWMHints(xDisplay,theWindowElement->xWindow,&theHints);

					theWindowElement->graphicsContext=XCreateGC(xDisplay,theWindowElement->xWindow,0,NULL);	// create context in which to draw the text (0 for mask) use default

					theWindowElement->invalidRegion=XCreateRegion();						// create new empty invalid region

					theWindowElement->currentIcon=(Pixmap)NULL;								// no icon associated with this yet

					theDocumentWindowInfo->verticalScrollBarOnLeft=verticalScrollBarOnLeft;	// copy global preference
					GetDocumentWindowRects(theWindow,&contentRect,&viewRect,&verticalRect,&horizontalRect,&statusRect);	// get sizes of areas within the window

					viewDescriptor.theBuffer=theBuffer;
					viewDescriptor.theRect=viewRect;										// place where the view is going
					viewDescriptor.topLine=0;
					viewDescriptor.leftPixel=0;
					viewDescriptor.tabSize=tabSize;

					viewDescriptor.backgroundColor=background;
					viewDescriptor.foregroundColor=foreground;

					viewDescriptor.fontName=fontName;
					viewDescriptor.active=true;												// create the view active
					viewDescriptor.viewTextChangedVector=AdjustDocumentWindowScrollThumbsAndStatus;		// fix the thumbs, and status when the view text changes
					viewDescriptor.viewPositionChangedVector=AdjustDocumentWindowScrollThumbs;			// fix the thumbs when the view position changes
					viewDescriptor.viewSelectionChangedVector=AdjustDocumentWindowViewStatus;			// fix the status when the selection changes
					if((theDocumentWindowInfo->theView=CreateEditorView(theWindow,&viewDescriptor)))	// create a view for this area, set up its font stuff
					{
						scrollDescriptor.x=verticalRect.x;									// set up description of the scroll bar
						scrollDescriptor.y=verticalRect.y;
						scrollDescriptor.length=verticalRect.h;
						scrollDescriptor.horizontal=false;
						scrollDescriptor.thumbMax=0;
						scrollDescriptor.thumbPosition=0;
						scrollDescriptor.stepLowerProc=ScrollEditViewUpEvent;
						scrollDescriptor.stepHigherProc=ScrollEditViewDownEvent;
						scrollDescriptor.pageLowerProc=ScrollEditViewPageUpEvent;
						scrollDescriptor.pageHigherProc=ScrollEditViewPageDownEvent;
						scrollDescriptor.thumbProc=ScrollEditViewThumbVerticalEvent;
						scrollDescriptor.procParameters=theDocumentWindowInfo->theView;

						if((theDocumentWindowInfo->verticalScrollBar=CreateScrollBar(theWindow,&scrollDescriptor)))
						{
							scrollDescriptor.x=horizontalRect.x;
							scrollDescriptor.y=horizontalRect.y;
							scrollDescriptor.length=horizontalRect.w;
							scrollDescriptor.horizontal=true;
							scrollDescriptor.thumbMax=0;
							scrollDescriptor.thumbPosition=0;
							scrollDescriptor.stepLowerProc=ScrollEditViewLeftEvent;
							scrollDescriptor.stepHigherProc=ScrollEditViewRightEvent;
							scrollDescriptor.pageLowerProc=ScrollEditViewPageLeftEvent;
							scrollDescriptor.pageHigherProc=ScrollEditViewPageRightEvent;
							scrollDescriptor.thumbProc=ScrollEditViewThumbHorizontalEvent;
							scrollDescriptor.procParameters=theDocumentWindowInfo->theView;

							if((theDocumentWindowInfo->horizontalScrollBar=CreateScrollBar(theWindow,&scrollDescriptor)))
							{
								AdjustDocumentWindowScrollThumbsAndStatus(theDocumentWindowInfo->theView);		// force the thumbs and status to be up to date
								XMapRaised(xDisplay,theWindowElement->xWindow);						// map window to the display
								WaitUntilMapped(theWindow);
								return(theWindow);
							}
							DisposeScrollBar(theDocumentWindowInfo->verticalScrollBar);
						}
						DisposeEditorView(theDocumentWindowInfo->theView);
					}
					XDestroyRegion(theWindowElement->invalidRegion);					// destroy invalid region
					XFreeGC(xDisplay,theWindowElement->graphicsContext);				// if something went wrong, get rid of graphics context
					XDestroyWindow(xDisplay,theWindowElement->xWindow);					// get rid of the window
					WaitUntilDestroyed(theWindow);

					UnlinkWindowElement(theWindowElement);
					MDisposePtr(theWindowElement);
				}
				FreeFont(theDocumentWindowInfo->statusBarFont);
			}
			MDisposePtr(theDocumentWindowInfo);
		}
		MDisposePtr(theWindow);
	}
	ReportMessage("Failed to create window\n");
	return((EDITOR_WINDOW *)NULL);
}

// routines callable by the editor

void MinimizeDocumentWindow(EDITOR_WINDOW *theWindow)
// minimize the passed window
{
	MinimizeEditorWindow(theWindow);
}

void UnminimizeDocumentWindow(EDITOR_WINDOW *theWindow)
// unminimize the passed window
{
	UnminimizeEditorWindow(theWindow);
}

void SetTopDocumentWindow(EDITOR_WINDOW *theWindow)
// set the passed document window to the top of the display
{
	SetTopEditorWindow(theWindow);
}

EDITOR_WINDOW *GetActiveDocumentWindow()
// return the document window that has the user's attention, or NULL if there is none
// depending on the system, this may or may not return the top-most document window
// (For instance under Motif, the active document window may be the one with the
// cursor over it)
{
	WINDOW_LIST_ELEMENT
		*theWindowElement;
	Window
		xWindow;
	int
		focusRevert;

	XGetInputFocus(xDisplay,&xWindow,&focusRevert);			// find out who has the focus
	if((theWindowElement=LocateWindowElement(xWindow)))		// see if it is one of our windows
	{
		if(theWindowElement->theEditorWindow->windowType==EWT_DOCUMENT)
		{
			return(theWindowElement->theEditorWindow);		// if the window with focus is a document window, then return it
		}
	}
	return(GetTopEditorWindowType(EWT_DOCUMENT));			// otherwise, the active one is the top one
}

EDITOR_WINDOW *GetTopDocumentWindow()
// return the top-most document window, or NULL if there is none
// this indirectly uses the realTopWindow fields of the window elements
{
	return(GetTopEditorWindowType(EWT_DOCUMENT));
}

bool GetSortedDocumentWindowList(UINT32 *numElements,EDITOR_WINDOW ***theList)
// return a list of all document windows -- sorted top to bottom
// if there is a problem, SetError, and return false
// otherwise return true. If true is returned, theList must be
// MDisposePtr'ed by the caller (even if numElements is 0)
{
	return(GetSortedEditorWindowTypeList(EWT_DOCUMENT,numElements,theList));
}

EDITOR_VIEW *GetDocumentWindowCurrentView(EDITOR_WINDOW *theWindow)
// passed a document window, return the current view of the window
{
	return(((GUI_DOCUMENT_WINDOW_INFO *)theWindow->windowInfo)->theView);			// currently, document windows contain one view only
}

bool GetDocumentWindowTitle(EDITOR_WINDOW *theWindow,char *theTitle,UINT32 stringBytes)
// copy theWindow's title into theTitle
// if theWindow's title is longer than stringBytes, return false
{
	char
		*theName;

	if(XFetchName(xDisplay,((WINDOW_LIST_ELEMENT *)theWindow->userData)->xWindow,&theName))
	{
		strncpy(theTitle,theName,stringBytes);
		XFree(theName);
	}
	else
	{
		strncpy(theTitle,"no name",stringBytes);
	}
	if(stringBytes)
	{
		theTitle[stringBytes-1]='\0';
	}
	return(true);
}

bool SetDocumentWindowTitle(EDITOR_WINDOW *theWindow,char *theTitle)
// replace theTitle of theWindow
// if there is a problem, return false
{
	XStoreName(xDisplay,((WINDOW_LIST_ELEMENT *)theWindow->userData)->xWindow,theTitle);
	XSetIconName(xDisplay,((WINDOW_LIST_ELEMENT *)theWindow->userData)->xWindow,theTitle);
	return(true);
}

bool GetEditorDocumentWindowForegroundColor(EDITOR_WINDOW *theWindow,EDITOR_COLOR *foregroundColor)
// get the foreground color in use by theWindow
// if there is a problem, SetError, and return false
{
	GUI_DOCUMENT_WINDOW_INFO
		*theWindowInfo;

	theWindowInfo=(GUI_DOCUMENT_WINDOW_INFO *)theWindow->windowInfo;
	*foregroundColor=theWindowInfo->foregroundColor;
	return(true);
}

bool SetEditorDocumentWindowForegroundColor(EDITOR_WINDOW *theWindow,EDITOR_COLOR foregroundColor)
// set the foreground color to be used in theWindow
// if there is a problem, leave the color unchanged, set the error
// and return false
{
	SetDocumentForegroundColor(theWindow,foregroundColor);
	XSetWindowBorder(xDisplay,((WINDOW_LIST_ELEMENT *)theWindow->userData)->xWindow,((GUI_DOCUMENT_WINDOW_INFO *)theWindow->windowInfo)->foregroundColor);
	return(true);
}

bool GetEditorDocumentWindowBackgroundColor(EDITOR_WINDOW *theWindow,EDITOR_COLOR *backgroundColor)
// get the background color in use by theWindow
// if there is a problem, SetError, and return false
{
	GUI_DOCUMENT_WINDOW_INFO
		*theWindowInfo;

	theWindowInfo=(GUI_DOCUMENT_WINDOW_INFO *)theWindow->windowInfo;
	*backgroundColor=theWindowInfo->backgroundColor;
	return(true);
}

bool SetEditorDocumentWindowBackgroundColor(EDITOR_WINDOW *theWindow,EDITOR_COLOR backgroundColor)
// set the background color to be used in theWindow
// if there is a problem, leave the color unchanged, set the error
// and return false
{
	EDITOR_RECT
		contentRect,
		viewRect,
		verticalRect,
		horizontalRect,
		statusRect;

	SetDocumentBackgroundColor(theWindow,backgroundColor);
	XSetWindowBackground(xDisplay,((WINDOW_LIST_ELEMENT *)theWindow->userData)->xWindow,((GUI_DOCUMENT_WINDOW_INFO *)theWindow->windowInfo)->backgroundColor);
	XClearWindow(xDisplay,((WINDOW_LIST_ELEMENT *)theWindow->userData)->xWindow);
	GetDocumentWindowRects(theWindow,&contentRect,&viewRect,&verticalRect,&horizontalRect,&statusRect);
	InvalidateWindowRect(theWindow,&contentRect);			// invalidate the whole thing, because we cleared it all
	return(true);
}

bool GetEditorDocumentWindowRect(EDITOR_WINDOW *theWindow,EDITOR_RECT *theRect)
// get the rectangle of theWindow
// if there is a problem, SetError, and return false
// NOTE: once again, X windows makes this MUCH more of a pain than it needs
// to be. We have to get the attributes of our REAL top window for the x, and y
// and of our actual window for the width, and height... sheesh....
{
	XWindowAttributes
		theAttributes;

	XGetWindowAttributes(xDisplay,((WINDOW_LIST_ELEMENT *)theWindow->userData)->realTopWindow,&theAttributes);
	theRect->x=theAttributes.x;
	theRect->y=theAttributes.y;

	XGetWindowAttributes(xDisplay,((WINDOW_LIST_ELEMENT *)theWindow->userData)->xWindow,&theAttributes);
	theRect->w=theAttributes.width;
	theRect->h=theAttributes.height;

	return(true);
}

bool SetEditorDocumentWindowRect(EDITOR_WINDOW *theWindow,EDITOR_RECT *theRect)
// set theWindow to theRect
// if there is a problem, leave theRect unchanged, set the error
// and return false
{
	XMoveResizeWindow(xDisplay,((WINDOW_LIST_ELEMENT *)theWindow->userData)->xWindow,theRect->x,theRect->y,theRect->w,theRect->h);
	return(true);
}

EDITOR_WINDOW *OpenDocumentWindow(EDITOR_BUFFER *theBuffer,EDITOR_RECT *theRect,char *theTitle,char *fontName,UINT32 tabSize,EDITOR_COLOR foreground,EDITOR_COLOR background)
// open a document window onto theBuffer
// if there is a problem, undo anything that was done,
// set the error, then return NULL
{
	return(CreateDocumentWindow(theBuffer,theRect,theTitle,fontName,tabSize,foreground,background));
}

void CloseDocumentWindow(EDITOR_WINDOW *theWindow)
// close an open document window
{
	DisposeDocumentWindow(theWindow);
}
