// Low level stuff needed to manage views
// Copyright (C) 2000 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

#define	STAGING_BUFFER_SIZE		8192										// number of bytes used in the staging buffer (must be larger than STAGING_MAX_ADD)
#define	STAGING_MAX_ADD			256											// maximum number of bytes that will ever be added to the staging buffer in one call to GetCharacterEquiv

static const char hexConv[]="0123456789ABCDEF";								// used for speedy hex conversions (for display of non-ascii codes)

// when text is being built, or widths are being calculated, these globals
// are used, they allow for much more efficient creation of text because things
// are not being passed all over for each character
static UINT8
	gceStagingBuffer[STAGING_BUFFER_SIZE];									// this is where get character equivalent builds its output
static UINT32
	gceBufferIndex;															// offset into the staging buffer
static UINT32
	gceColumnNumber;														// running column number
static UINT16
	*gceWidthTable;															// width table to use while calculating positions
static UINT32
	gceTabSize;																// number of spaces to use per tab
static INT32
	gceRunningWidth;														// running width

static EDITOR_FONT *GetFontForStyle(VIEW_STYLE *theStyles,UINT32 desiredStyle)
// Return the font that should be used for the given style
// if there is none specified, then use the default
{
	if(theStyles[desiredStyle].theFont)							// if there is a font loaded for this style, return it
	{
		return(theStyles[desiredStyle].theFont);
	}
	return(theStyles[0].theFont);								// otherwise, return the default font
}

static EDITOR_COLOR GetForegroundForStyle(VIEW_STYLE *theStyles,UINT32 desiredStyle)
// Return the foreground color that should be used for the given style
// if there is none specified, then use the default
{
	if(theStyles[desiredStyle].haveForegroundColor)					// if there is a color loaded for this style, return it
	{
		return(theStyles[desiredStyle].foregroundColor);
	}
	return(theStyles[0].foregroundColor);							// otherwise, return the default color
}

static EDITOR_COLOR GetBackgroundForStyle(VIEW_STYLE *theStyles,UINT32 desiredStyle)
// Return the background color that should be used for the given style
// if there is none specified, then use the default
{
	if(theStyles[desiredStyle].haveBackgroundColor)					// if there is a color loaded for this style, return it
	{
		return(theStyles[desiredStyle].backgroundColor);
	}
	return(theStyles[0].backgroundColor);							// otherwise, return the default color
}

static bool PointInView(EDITOR_VIEW *theView,INT32 x,INT32 y)
// Return true if parent window relative x/y are in theView
// false if not
{
	return(PointInRECT(x,y,&(theView->viewInfo->bounds)));
}

static void LocalClickToViewClick(EDITOR_VIEW *theView,INT32 clickX,INT32 clickY,INT32 *xPosition,INT32 *lineNumber)
// Given a point local to the window that contains theView,
// convert it into a view position
{
	GUI_VIEW_INFO
		*theViewInfo;
	INT32
		viewY;

	theViewInfo=theView->viewInfo;								// point at the information record for this view
	(*xPosition)=clickX-theViewInfo->textBounds.x;
	viewY=clickY-theViewInfo->textBounds.y;
	if(viewY<0)
	{
		viewY-=theViewInfo->lineHeight;							// account for cases about 0
	}
	(*lineNumber)=viewY/((INT32)(theViewInfo->lineHeight));
}

void HandleViewKeyEvent(EDITOR_VIEW *theView,XEvent *theEvent)
// Handle an incoming key event for theView
{
	VIEW_EVENT
		theViewEvent;
	EDITOR_KEY
		theKeyData;

	theViewEvent.theView=theView;							// point at the view
	theViewEvent.eventType=VET_KEYDOWN;						// let it know what type
	if(KeyEventToEditorKey(theEvent,&theKeyData))			// fill in the key record
	{
		theViewEvent.eventData=&theKeyData;
		HandleViewEvent(&theViewEvent);						// pass this off to high level handler
	}
}

static bool GetCharacterEquiv(UINT8 theChar)
// Put the correct characters into the gceStagingBuffer at gceBufferIndex
// update gceColumnNumber, gceBufferIndex, and gceRunningWidth
// if an end of line character is seen, return false, else return true
// NOTE: this is allowed to add at most STAGING_MAX_ADD characters to the staging buffer
// When called, there must be at least that many bytes available, or
// there will be trouble
{
	UINT32
		i,
		numTabSpaces;

	if(theChar>=' '&&gceWidthTable[theChar])				// make sure the font can display this, if not, display it as hex
	{
		gceStagingBuffer[gceBufferIndex++]=theChar;
		gceRunningWidth+=gceWidthTable[theChar];
		gceColumnNumber++;
	}
	else
	{
		switch(theChar)
		{
			case '\n':
				return(false);
				break;
			case '\t':
				if(gceTabSize)
				{
					numTabSpaces=gceTabSize-(gceColumnNumber%gceTabSize);
					for(i=0;i<numTabSpaces;i++)
					{
						gceStagingBuffer[gceBufferIndex++]=' ';			// space over to tab stop
					}
					gceRunningWidth+=gceWidthTable[' ']*numTabSpaces;	// add in space to running width
					gceColumnNumber+=numTabSpaces;						// push to correct column
				}
				else
				{
					gceRunningWidth+=gceWidthTable[gceStagingBuffer[gceBufferIndex++]='['];
					gceRunningWidth+=gceWidthTable[gceStagingBuffer[gceBufferIndex++]=hexConv[theChar>>4]];
					gceRunningWidth+=gceWidthTable[gceStagingBuffer[gceBufferIndex++]=hexConv[theChar&0x0F]];
					gceRunningWidth+=gceWidthTable[gceStagingBuffer[gceBufferIndex++]=']'];
					gceColumnNumber+=4;
				}
				break;
			default:
				// this is a little scary looking, but it is fast
				gceRunningWidth+=gceWidthTable[gceStagingBuffer[gceBufferIndex++]='['];
				gceRunningWidth+=gceWidthTable[gceStagingBuffer[gceBufferIndex++]=hexConv[theChar>>4]];
				gceRunningWidth+=gceWidthTable[gceStagingBuffer[gceBufferIndex++]=hexConv[theChar&0x0F]];
				gceRunningWidth+=gceWidthTable[gceStagingBuffer[gceBufferIndex++]=']'];
				gceColumnNumber+=4;
				break;
		}
	}
	return(true);
}

static bool PositionAtPixel(EDITOR_VIEW *theView,UINT32 linePosition,INT32 xPosition,UINT32 *charPosition,UINT32 *charColumn,INT32 *charStartPixel,UINT32 *charWidth)
// Given linePosition which points to the character at the start of a line, and xPosition
// as a "pixel" position along the line, determine the offset to the character
// at xPosition, the pixel where that character begins, and the width of that character.
// if linePosition is past EOF, charPosition will be returned as linePosition,
// charColumn will be 0, charStartPixel will be VIEW_LEFT_MARGIN, and charWidth will be 0
// if the xPosition is past the end of the line, or end of the data, this will return false
{
	GUI_VIEW_INFO
		*theViewInfo;
	CHUNK_HEADER
		*theChunk;
	UINT32
		theOffset;
	bool
		haveStyle,
		inLine,
		loopDone;
	INT32
		lastRunningWidth;
	UINT32
		lastColumnNumber;
	UINT32
		stylePosition,
		numStyleElements,
		currentStyle;

	theViewInfo=theView->viewInfo;							// point at the information record for this view

	PositionToChunkPosition(theView->parentBuffer->textUniverse,linePosition,&theChunk,&theOffset);

	if((haveStyle=GetStyleRange(theView->parentBuffer->styleUniverse,linePosition,&stylePosition,&numStyleElements,&currentStyle)))
	{
		ForceStyleIntoRange(currentStyle);
		stylePosition+=numStyleElements;					// style will change at this position
	}

	loopDone=false;
	inLine=true;

	gceTabSize=theView->tabSize;
	gceRunningWidth=lastRunningWidth=VIEW_LEFT_MARGIN;
	gceColumnNumber=lastColumnNumber=0;
	gceWidthTable=&(GetFontForStyle(theView->viewStyles,currentStyle)->charWidths[0]);

	while(!loopDone&&inLine&&theChunk)						// run out to the position where the text begins
	{
		if(haveStyle&&(linePosition>=stylePosition))		// time to update style?
		{
			if((haveStyle=GetStyleRange(theView->parentBuffer->styleUniverse,linePosition,&stylePosition,&numStyleElements,&currentStyle)))
			{
				ForceStyleIntoRange(currentStyle);
				stylePosition+=numStyleElements;			// style will change at this position
			}
			gceWidthTable=&(GetFontForStyle(theView->viewStyles,currentStyle)->charWidths[0]);
		}
		gceBufferIndex=0;
		if((inLine=GetCharacterEquiv(theChunk->data[theOffset])))
		{
			if(gceRunningWidth>xPosition)
			{
				loopDone=true;
			}
			else
			{
				lastColumnNumber=gceColumnNumber;			// remember these
				lastRunningWidth=gceRunningWidth;
				linePosition++;
				theOffset++;
				if(theOffset>=theChunk->totalBytes)			// see if time for a new chunk
				{
					theChunk=theChunk->nextHeader;
					theOffset=0;
				}
			}
		}
	}
	*charPosition=linePosition;								// return the results
	*charColumn=lastColumnNumber;
	*charStartPixel=lastRunningWidth;
	*charWidth=gceRunningWidth-lastRunningWidth;

	return(theChunk&&inLine);
}

static void PixelAtPosition(EDITOR_VIEW *theView,UINT32 thePosition,INT32 *xPosition,bool limitMax)
// Given thePosition within theView, determine the "pixel" offset to the start of the character
// of the line thePosition is in.
// if thePosition is passed in past the end of the buffer, this will return 0
// NOTE: if limitMax is true, this will abort as soon as the returned position
// grows larger than that passed in
{
	GUI_VIEW_INFO
		*theViewInfo;
	CHUNK_HEADER
		*theChunk;
	UINT32
		theOffset;
	UINT32
		linePosition;
	UINT32
		theLine,
		theLineOffset;
	bool
		haveStyle;
	UINT32
		stylePosition,
		numStyleElements,
		currentStyle;

	theViewInfo=theView->viewInfo;							// point at the information record for this view

	PositionToLinePosition(theView->parentBuffer->textUniverse,thePosition,&theLine,&theLineOffset,&theChunk,&theOffset);

	linePosition=thePosition-theLineOffset;

	if((haveStyle=GetStyleRange(theView->parentBuffer->styleUniverse,linePosition,&stylePosition,&numStyleElements,&currentStyle)))
	{
		ForceStyleIntoRange(currentStyle);
		stylePosition+=numStyleElements;					// style will change at this position
	}

	gceTabSize=theView->tabSize;
	gceRunningWidth=VIEW_LEFT_MARGIN;
	gceColumnNumber=0;
	gceWidthTable=&(GetFontForStyle(theView->viewStyles,currentStyle)->charWidths[0]);

	while((linePosition<thePosition)&&theChunk&&(!limitMax||(gceRunningWidth<*xPosition)))	// run out to the given position
	{
		if(haveStyle&&(linePosition>=stylePosition))		// time to update style?
		{
			if((haveStyle=GetStyleRange(theView->parentBuffer->styleUniverse,linePosition,&stylePosition,&numStyleElements,&currentStyle)))
			{
				ForceStyleIntoRange(currentStyle);
				stylePosition+=numStyleElements;			// style will change at this position
			}
			gceWidthTable=&(GetFontForStyle(theView->viewStyles,currentStyle)->charWidths[0]);
		}
		gceBufferIndex=0;
		GetCharacterEquiv(theChunk->data[theOffset]);
		linePosition++;
		theOffset++;
		if(theOffset>=theChunk->totalBytes)					// see if time for a new chunk
		{
			theChunk=theChunk->nextHeader;
			theOffset=0;
		}
	}
	(*xPosition)=gceRunningWidth;
}

void GetEditorViewGraphicToTextPosition(EDITOR_VIEW *theView,UINT32 linePosition,INT32 xPosition,UINT32 *betweenOffset,UINT32 *charOffset)
// Given linePosition which points to the character at the start of a line, and xPosition
// as a "pixel" position along the line, determine two offsets:
// betweenOffset is the offset of the character boundary closest to the xPosition,
// charOffset is the offset of the character closest to the xPosition
// if linePosition is past EOF, both offsets will be returned as 0
{
	UINT32
		charPosition;
	UINT32
		charColumn;
	INT32
		charStartPixel;
	UINT32
		charWidth;
	bool
		inLine;

	inLine=PositionAtPixel(theView,linePosition,xPosition,&charPosition,&charColumn,&charStartPixel,&charWidth);
	*betweenOffset=*charOffset=charPosition-linePosition;
	if(inLine&&(xPosition>(charStartPixel+((INT32)charWidth/2))))
	{
		(*betweenOffset)++;
	}
}

void GetEditorViewTextToGraphicPosition(EDITOR_VIEW *theView,UINT32 thePosition,INT32 *xPosition,bool limitMax,UINT32 *slopLeft,UINT32 *slopRight)
// Given thePosition within theView, determine the "pixel" offset to the start of the character
// of the line thePosition is in.
// if thePosition is passed in past the end of the buffer, this will return the width of the last line
// NOTE: if limitMax is true, xPosition arrives at this function with a maximum
// pixel position. This routine is then allowed to stop as soon as it
// sees the return result-slopLeft as greater than the passed value.
// NOTE: slopLeft and slopRight are used to allow fonts which have overhang
// into the previous or next character to be dealt with correctly.
// slopLeft tells how far the character at thePosition draws to the left of
// its given position.
// slopRight tells how far the character 1 to the left of thePosition draws
// to the right (past its character width).
{
	GUI_VIEW_INFO
		*theViewInfo;

	theViewInfo=theView->viewInfo;
	*slopLeft=theViewInfo->maxLeftOverhang;
	*slopRight=theViewInfo->maxRightOverhang;
	if(limitMax)					// if limiting to pixel position, move out by max overhang
	{
		*xPosition+=*slopLeft;
	}
	PixelAtPosition(theView,thePosition,xPosition,limitMax);
}

void GetEditorViewTextInfo(EDITOR_VIEW *theView,UINT32 *topLine,UINT32 *numLines,INT32 *leftPixel,UINT32 *numPixels)
// Get information about the given view
{
	GUI_VIEW_INFO
		*theViewInfo;

	theViewInfo=theView->viewInfo;
	*topLine=theViewInfo->topLine;
	*numLines=theViewInfo->numLines;
	*leftPixel=theViewInfo->leftPixel;
	*numPixels=theViewInfo->textBounds.w;
}

static void InvertViewCursor(EDITOR_VIEW *theView)
// Invert the cursor wherever it happens to be in theView
// theView's window's current graphics context will be used to constrain the draw
{
	GUI_VIEW_INFO
		*theViewInfo;
	CHUNK_HEADER
		*theChunk;
	UINT32
		theOffset;
	UINT32
		theLine,
		theLineOffset;
	INT32
		xPosition,
		maxXPosition;
	UINT32
		slopLeft,
		slopRight;
	EDITOR_RECT
		cursorRect;
	Window
		xWindow;
	GC
		graphicsContext;
	UINT32
		theStyle;

	theViewInfo=theView->viewInfo;									// point at the information record for this view

	PositionToLinePosition(theView->parentBuffer->textUniverse,GetSelectionCursorPosition(theView->parentBuffer->selectionUniverse),&theLine,&theLineOffset,&theChunk,&theOffset);
	if((theLine>=theViewInfo->topLine)&&(theLine<theViewInfo->topLine+theViewInfo->numLines))	// make sure cursor is on the view
	{
		maxXPosition=xPosition=theViewInfo->leftPixel+theViewInfo->textBounds.w;
		GetEditorViewTextToGraphicPosition(theView,GetSelectionCursorPosition(theView->parentBuffer->selectionUniverse),&xPosition,true,&slopLeft,&slopRight);

		theStyle=0;													// always paint the cursor in style 0

		if((xPosition>=theViewInfo->leftPixel)&&(xPosition<maxXPosition))	// see if on view horizontally
		{
			cursorRect.x=theViewInfo->textBounds.x+(xPosition-theViewInfo->leftPixel);
			cursorRect.w=1;
			cursorRect.y=theViewInfo->textBounds.y+theViewInfo->lineHeight*(theLine-theViewInfo->topLine);
			cursorRect.h=theViewInfo->lineHeight;

			xWindow=((WINDOW_LIST_ELEMENT *)theView->parentWindow->userData)->xWindow;		// get x window for this view
			graphicsContext=((WINDOW_LIST_ELEMENT *)theView->parentWindow->userData)->graphicsContext;	// get graphics context for this window

			XSetFunction(xDisplay,graphicsContext,GXxor);			// prepare to fill in xor format
			XSetFillStyle(xDisplay,graphicsContext,FillSolid);		// use solid fill

			XSetForeground(xDisplay,graphicsContext,EditorColorToXPixel(GetForegroundForStyle(theView->viewStyles,theStyle)^GetBackgroundForStyle(theView->viewStyles,theStyle)));	// this is the value to xor with

			XFillRectangle(xDisplay,xWindow,graphicsContext,cursorRect.x,cursorRect.y,cursorRect.w,cursorRect.h);

			XSetFunction(xDisplay,graphicsContext,GXcopy);			// put function back
		}
	}
}

static bool CursorShowing(EDITOR_VIEW *theView)
// Returns true if the cursor is showing
{
	GUI_VIEW_INFO
		*theViewInfo;

	theViewInfo=theView->viewInfo;							// point at the information record for this view
	if((IsSelectionEmpty(theView->parentBuffer->selectionUniverse))&&theViewInfo->cursorOn&&theViewInfo->blinkState&&theView->active)
	{
		return(true);
	}
	return(false);
}

static void DrawViewCursor(EDITOR_VIEW *theView)
// Look at the cursor state for this view, and draw it as needed
// it will be drawn into the current port, and current clip region
{
	if(CursorShowing(theView))
	{
		InvertViewCursor(theView);							// if it is showing, then draw it
	}
}

static void AddSelectionLineToRegion(EDITOR_VIEW *theView,SELECTION_UNIVERSE *theSelection,Region theRegion,UINT32 currentPosition,INT32 yWindowOffset)
// Read bytes from the line of text at theChunk, theOffset
// Add pieces where the passed selection is active to theRegion
// NOTE: this must work as an exact mirror to DrawViewLine
// NOTE: this adds extra space to the left and right of the region
// as it creates it, to accommodate the overhang of fonts
{
	GUI_VIEW_INFO
		*theViewInfo;
	CHUNK_HEADER
		*theChunk;
	UINT32
		theOffset;
	INT32
		xWindowOffset,
		xWindowEndOffset;
	INT32
		skippedWidth;
	bool
		currentSelection,
		nextSelection;
	UINT32
		currentStyle,
		nextStyle;
	INT32
		scrollX;					// pixel offset in text where we need to start drawing
	INT32
		textXOffset;				// window relative pixel offset where text is being drawn
	INT32
		distanceToStart;
	XRectangle
		selectionRect;
	UINT32
		charWidth;

	theViewInfo=theView->viewInfo;							// point at the information record for this view

	xWindowOffset=theViewInfo->textBounds.x;
	xWindowEndOffset=theViewInfo->textBounds.x+theViewInfo->textBounds.w;

	selectionRect.y=yWindowOffset;
	selectionRect.height=theViewInfo->lineHeight;

	scrollX=theViewInfo->leftPixel;							// get pixel offset in text where we need to start drawing

	PositionAtPixel(theView,currentPosition,scrollX,&currentPosition,&gceColumnNumber,&skippedWidth,&charWidth);

	PositionToChunkPosition(theView->parentBuffer->textUniverse,currentPosition,&theChunk,&theOffset);

	gceBufferIndex=0;										// begin at start of buffer
	gceRunningWidth=0;

	distanceToStart=skippedWidth-scrollX;

	if(distanceToStart>(xWindowEndOffset-xWindowOffset))	// do not go past the end of the draw area
	{
		distanceToStart=xWindowEndOffset-xWindowOffset;
	}

	textXOffset=xWindowOffset+distanceToStart;				// move the offset slightly back to beginning of first character which enters the view (or forward if view is scrolled negatively)

	currentSelection=GetSelectionAtPosition(theSelection,currentPosition);
	currentStyle=GetStyleAtPosition(theView->parentBuffer->styleUniverse,currentPosition);
	ForceStyleIntoRange(currentStyle);

	gceWidthTable=&(GetFontForStyle(theView->viewStyles,currentStyle)->charWidths[0]);

	while(theChunk&&((textXOffset+gceRunningWidth)<xWindowEndOffset)&&GetCharacterEquiv(theChunk->data[theOffset]))		// run through the text
	{
		nextSelection=GetSelectionAtPosition(theSelection,currentPosition+1);

		nextStyle=GetStyleAtPosition(theView->parentBuffer->styleUniverse,currentPosition+1);
		ForceStyleIntoRange(nextStyle);

		if((currentStyle!=nextStyle)||(currentSelection!=nextSelection)||(gceBufferIndex>STAGING_BUFFER_SIZE-STAGING_MAX_ADD))	// is something about to change? (if so, output what we have so far) (this will also output if the staging buffer is getting full)
		{
			if(currentSelection)
			{
				selectionRect.x=textXOffset-theViewInfo->maxLeftOverhang;
				selectionRect.width=gceRunningWidth+theViewInfo->maxLeftOverhang+theViewInfo->maxRightOverhang;
				if(selectionRect.x<xWindowOffset)			// pin to bounds
				{
					selectionRect.width-=(xWindowOffset-selectionRect.x);
					selectionRect.x=xWindowOffset;
				}
				if(selectionRect.x+selectionRect.width>xWindowEndOffset)
				{
					selectionRect.width=xWindowEndOffset-selectionRect.x;
				}
				XUnionRectWithRegion(&selectionRect,theRegion,theRegion);
			}

			textXOffset+=gceRunningWidth;

			gceBufferIndex=0;								// begin at start of buffer
			gceRunningWidth=0;

			currentSelection=nextSelection;
			currentStyle=nextStyle;
			gceWidthTable=&(GetFontForStyle(theView->viewStyles,currentStyle)->charWidths[0]);
		}

		currentPosition++;
		theOffset++;
		if(theOffset>=theChunk->totalBytes)					// push through end of chunk
		{
			theChunk=theChunk->nextHeader;
			theOffset=0;
		}
	}
	// ran off the end of the line, or the end of the universe
	if(gceBufferIndex)
	{
		if(currentSelection)
		{
			selectionRect.x=textXOffset-theViewInfo->maxLeftOverhang;
			selectionRect.width=gceRunningWidth+theViewInfo->maxLeftOverhang+theViewInfo->maxRightOverhang;
			if(selectionRect.x<xWindowOffset)					// pin to bounds
			{
				selectionRect.width-=(xWindowOffset-selectionRect.x);
				selectionRect.x=xWindowOffset;
			}
			if(selectionRect.x+selectionRect.width>xWindowEndOffset)
			{
				selectionRect.width=xWindowEndOffset-selectionRect.x;
			}
			XUnionRectWithRegion(&selectionRect,theRegion,theRegion);
		}
		textXOffset+=gceRunningWidth;
	}

	if(textXOffset<xWindowEndOffset)							// do we need to go out to the end of the line?
	{
		if(textXOffset<xWindowOffset)							// did this end before the draw area began?
		{
			textXOffset=xWindowOffset;
		}
		currentSelection=GetSelectionAtPosition(theSelection,currentPosition);		// get selection information for newline (or off of end)

		if(currentSelection)
		{
			selectionRect.x=textXOffset;
			selectionRect.width=xWindowEndOffset-textXOffset;
			if(selectionRect.x<xWindowOffset)					// pin to bounds
			{
				selectionRect.width-=(xWindowOffset-selectionRect.x);
				selectionRect.x=xWindowOffset;
			}
			if(selectionRect.x+selectionRect.width>xWindowEndOffset)
			{
				selectionRect.width=xWindowEndOffset-selectionRect.x;
			}
			XUnionRectWithRegion(&selectionRect,theRegion,theRegion);
		}
	}
}

static EDITOR_COLOR SumColors(EDITOR_COLOR colorA,EDITOR_COLOR colorB)
// When marks overlap, this is called to blend the colors of the overlapping
// marks together
{
	unsigned int
		red,
		green,
		blue;

	red=((colorA>>16)&0xFF)+((colorB>>16)&0xFF);
	green=((colorA>>8)&0xFF)+((colorB>>8)&0xFF);
	blue=(colorA&0xFF)+(colorB&0xFF);
	if(red>255)
	{
		red=255;
	}
	if(green>255)
	{
		green=255;
	}
	if(blue>255)
	{
		blue=255;
	}
	return((red<<16)|(green<<8)|blue);
}

static bool DetermineMarksBackgroundColorAtPosition(EDITOR_VIEW *theView,UINT32 position,EDITOR_COLOR *color)
// Work out what the mark background color contribution is for the given position
// This looks at all visible marks, and sums their individual background color
// components. If there are no visible marks at the position, false is returned
{
	bool
		sawMark;
	MARK_VISIBILITY
		*visibleMark;

	sawMark=false;
	visibleMark=theView->headVisibility;				// get head of visible marks
	while(visibleMark)
	{
		if(GetSelectionAtPosition(visibleMark->parentMark->selectionUniverse,position))	// mark active?
		{
			if(!sawMark)								// if first mark, just copy the color
			{
				*color=visibleMark->backgroundColor;
				sawMark=true;
			}
			else										// sum in other colors
			{
				*color=SumColors(*color,visibleMark->backgroundColor);
			}
		}
		visibleMark=visibleMark->nextMark;				// move to next mark hanging off of this view
	}
	return(sawMark);
}

static void DrawViewBackgroundPiece(EDITOR_VIEW *theView,bool selected,bool haveMarkColor,EDITOR_COLOR markColor,UINT32 theStyle,INT32 xWindowOffset,UINT32 width,INT32 yWindowOffset)
// Draw the background into the view in the given style and selection
{
	Window
		xWindow;
	GC
		graphicsContext;
	GUI_VIEW_INFO
		*theViewInfo;
	EDITOR_COLOR
		background;

	theViewInfo=theView->viewInfo;						// point at the information record for this view
	graphicsContext=((WINDOW_LIST_ELEMENT *)theView->parentWindow->userData)->graphicsContext;	// get graphics context for this window
	xWindow=((WINDOW_LIST_ELEMENT *)theView->parentWindow->userData)->xWindow;					// get x window for this window

	if(selected)
	{
		if(!theView->haveSelectionForegroundColor&&!theView->haveSelectionBackgroundColor)		// if neither specified, invert
		{
			background=GetForegroundForStyle(theView->viewStyles,theStyle);
		}
		else	// otherwise, do not invert
		{
			if(theView->haveSelectionBackgroundColor)
			{
				background=theView->selectionBackgroundColor;
			}
			else
			{
				background=GetBackgroundForStyle(theView->viewStyles,theStyle);
			}
		}
	}
	else
	{
		if(!haveMarkColor)
		{
			background=GetBackgroundForStyle(theView->viewStyles,theStyle);
		}
		else
		{
			background=markColor;
		}
	}


	XSetForeground(xDisplay,graphicsContext,EditorColorToXPixel(background));	// set fill color

	XFillRectangle(xDisplay,xWindow,graphicsContext,xWindowOffset,yWindowOffset,width,theViewInfo->lineHeight);
}

static void DrawViewTextPiece(EDITOR_VIEW *theView,bool selected,UINT32 theStyle,INT32 xWindowOffset,UINT32 width,INT32 yWindowOffset,INT32 clipOffset)
// Draw the piece of text in the staging buffer into the view in the given style and selection
{
	Window
		xWindow;
	GC
		graphicsContext;
	GUI_VIEW_INFO
		*theViewInfo;
	EDITOR_COLOR
		foreground;

	theViewInfo=theView->viewInfo;						// point at the information record for this view
	graphicsContext=((WINDOW_LIST_ELEMENT *)theView->parentWindow->userData)->graphicsContext;	// get graphics context for this window
	xWindow=((WINDOW_LIST_ELEMENT *)theView->parentWindow->userData)->xWindow;					// get x window for this window

	XSetFont(xDisplay,graphicsContext,GetFontForStyle(theView->viewStyles,theStyle)->theXFont->fid);	// point to the font needed for this draw
	if(selected)
	{
		if(!theView->haveSelectionForegroundColor&&!theView->haveSelectionBackgroundColor)		// if neither specified, invert
		{
			foreground=GetBackgroundForStyle(theView->viewStyles,theStyle);
		}
		else	// otherwise, do not invert
		{
			if(theView->haveSelectionForegroundColor)
			{
				foreground=theView->selectionForegroundColor;
			}
			else
			{
				foreground=GetForegroundForStyle(theView->viewStyles,theStyle);
			}
		}
	}
	else
	{
		foreground=GetForegroundForStyle(theView->viewStyles,theStyle);
	}

	XSetForeground(xDisplay,graphicsContext,EditorColorToXPixel(foreground));

// code that would be here if there was no hack
//	XDrawString(xDisplay,xWindow,graphicsContext,xWindowOffset,yWindowOffset+theViewInfo->lineAscent,(char *)gceStagingBuffer,gceBufferIndex);

//------------------------------------------------------------------------------------------------------------------
// NOTE: this hack brought to you by X windows....
// If XDrawString has to clip the text at the right boundary, it draws much slower than
// if it does not. So.... we detect that case and make 2 calls to draw (which are usually much faster
// than 1)
// The first call draws the entire line up to the first non-clipping character
// The second call just draws the clipping character
{
	UINT32
		charactersBack,
		pixelsBack;

	if(xWindowOffset+(INT32)width>clipOffset)
	{
		charactersBack=0;
		pixelsBack=0;
		while((charactersBack<gceBufferIndex)&&((xWindowOffset+(INT32)(width-pixelsBack))>clipOffset))
		{
			charactersBack++;
			pixelsBack+=gceWidthTable[gceStagingBuffer[gceBufferIndex-charactersBack]];
		}
		XDrawString(xDisplay,xWindow,graphicsContext,xWindowOffset,yWindowOffset+theViewInfo->lineAscent,(char *)gceStagingBuffer,gceBufferIndex-charactersBack);
		XDrawString(xDisplay,xWindow,graphicsContext,xWindowOffset+width-pixelsBack,yWindowOffset+theViewInfo->lineAscent,(char *)&gceStagingBuffer[gceBufferIndex-charactersBack],charactersBack);
	}
	else	// no problem, the text is not going to clip
	{
		XDrawString(xDisplay,xWindow,graphicsContext,xWindowOffset,yWindowOffset+theViewInfo->lineAscent,(char *)gceStagingBuffer,gceBufferIndex);
	}
}
//------------------------------------------------------------------------------------------------------------------
}

static void DrawViewLine(EDITOR_VIEW *theView,UINT32 currentPosition,INT32 xWindowOffset,INT32 xWindowEndOffset,INT32 yWindowOffset)
// Read bytes from the line of text at currentPosition
// Draw them onto the view
// This is responsible for drawing ALL of the pixels in the view for the line
// on which it was called (even if that line contains no text)
// NOTE: this performs the draw in 2 passes.
// The first pass fills the background for the line, and the second
// pass draws the text over the background.
// The draw is done this way so that fonts which have pixels which
// overhang their character cells will draw without being clipped off
{
	CHUNK_HEADER
		*theChunk;
	UINT32
		theOffset;
	Window
		xWindow;
	GUI_VIEW_INFO
		*theViewInfo;
	INT32
		skippedWidth;
	GC
		graphicsContext;
	UINT32
		currentStyle,
		nextStyle;
	bool
		currentSelection,
		nextSelection,
		currentMark,
		nextMark;
	EDITOR_COLOR
		currentMarkColor,
		nextMarkColor;
	INT32
		scrollX;					// pixel offset in text where we need to start drawing
	INT32
		textXOffset;				// window relative pixel offset where text is being drawn
	INT32
		distanceToStart;
	UINT32
		charWidth;

	UINT32
		beginningPosition;
	CHUNK_HEADER
		*beginningChunk;
	UINT32
		beginningOffset;
	UINT32
		beginningColumn;

	theViewInfo=theView->viewInfo;	// point at the information record for this view
	graphicsContext=((WINDOW_LIST_ELEMENT *)theView->parentWindow->userData)->graphicsContext;	// get graphics context for this window
	xWindow=((WINDOW_LIST_ELEMENT *)theView->parentWindow->userData)->xWindow;					// get x window for this window

	scrollX=theViewInfo->leftPixel+(xWindowOffset-theViewInfo->textBounds.x);	// get pixel offset in text where we need to start drawing

	PositionAtPixel(theView,currentPosition,scrollX,&currentPosition,&gceColumnNumber,&skippedWidth,&charWidth);

	PositionToChunkPosition(theView->parentBuffer->textUniverse,currentPosition,&theChunk,&theOffset);

	beginningPosition=currentPosition;				// remember these for later
	beginningChunk=theChunk;
	beginningOffset=theOffset;
	beginningColumn=gceColumnNumber;

	gceBufferIndex=0;								// begin at start of buffer
	gceRunningWidth=0;

	distanceToStart=skippedWidth-scrollX;

	if(distanceToStart>0)							// blank space before text starts?
	{
		if(distanceToStart>(xWindowEndOffset-xWindowOffset))	// do not go past the end of the draw area
		{
			distanceToStart=xWindowEndOffset-xWindowOffset;
		}
		XSetFillStyle(xDisplay,graphicsContext,FillSolid);
		XSetForeground(xDisplay,graphicsContext,EditorColorToXPixel(theView->viewStyles[0].backgroundColor));
		XFillRectangle(xDisplay,xWindow,graphicsContext,xWindowOffset,yWindowOffset,distanceToStart,theViewInfo->lineHeight);	// fill with background color
	}

	// first draw all background

	textXOffset=xWindowOffset+distanceToStart;		// move the offset slightly back to beginning of first character which enters the view (or forward if view is scrolled negatively)

	currentSelection=GetSelectionAtPosition(theView->parentBuffer->selectionUniverse,currentPosition);
	currentMark=DetermineMarksBackgroundColorAtPosition(theView,currentPosition,&currentMarkColor);
	currentStyle=GetStyleAtPosition(theView->parentBuffer->styleUniverse,currentPosition);
	ForceStyleIntoRange(currentStyle);

	gceWidthTable=&(GetFontForStyle(theView->viewStyles,currentStyle)->charWidths[0]);

	while(theChunk&&((textXOffset+gceRunningWidth)<xWindowEndOffset)&&GetCharacterEquiv(theChunk->data[theOffset]))		// fill in all of the background needed
	{
		nextSelection=GetSelectionAtPosition(theView->parentBuffer->selectionUniverse,currentPosition+1);
		nextMark=DetermineMarksBackgroundColorAtPosition(theView,currentPosition+1,&nextMarkColor);
		nextStyle=GetStyleAtPosition(theView->parentBuffer->styleUniverse,currentPosition+1);
		ForceStyleIntoRange(nextStyle);

		if((currentStyle!=nextStyle)||(currentSelection!=nextSelection)||(currentMark!=nextMark)||(currentMark&&(currentMarkColor!=nextMarkColor))||(gceBufferIndex>STAGING_BUFFER_SIZE-STAGING_MAX_ADD))	// is something about to change? (if so, output background so far) (this will also output if the staging buffer is getting full)
		{
			DrawViewBackgroundPiece(theView,currentSelection,currentMark,currentMarkColor,currentStyle,textXOffset,gceRunningWidth,yWindowOffset);

			textXOffset+=gceRunningWidth;

			gceBufferIndex=0;						// begin at start of buffer
			gceRunningWidth=0;

			currentSelection=nextSelection;
			currentMark=nextMark;
			currentMarkColor=nextMarkColor;
			currentStyle=nextStyle;
			gceWidthTable=&(GetFontForStyle(theView->viewStyles,currentStyle)->charWidths[0]);
		}

		currentPosition++;
		theOffset++;
		if(theOffset>=theChunk->totalBytes)			// push through end of chunk
		{
			theChunk=theChunk->nextHeader;
			theOffset=0;
		}
	}

	// ran off the end of the line, or the end of the universe
	if(gceBufferIndex)
	{
		DrawViewBackgroundPiece(theView,currentSelection,currentMark,currentMarkColor,currentStyle,textXOffset,gceRunningWidth,yWindowOffset);
		textXOffset+=gceRunningWidth;
	}

	if(textXOffset<xWindowEndOffset)				// do we need to draw out to the end of the line?
	{
		if(textXOffset<xWindowOffset)				// did this end before the draw area began?
		{
			textXOffset=xWindowOffset;
		}
		currentSelection=GetSelectionAtPosition(theView->parentBuffer->selectionUniverse,currentPosition);
		currentMark=DetermineMarksBackgroundColorAtPosition(theView,currentPosition,&currentMarkColor);
		currentStyle=GetStyleAtPosition(theView->parentBuffer->styleUniverse,currentPosition);	// get style and selection information for newline (or off of end)
		ForceStyleIntoRange(currentStyle);

		DrawViewBackgroundPiece(theView,currentSelection,currentMark,currentMarkColor,currentStyle,textXOffset,xWindowEndOffset-textXOffset,yWindowOffset);
	}


	// next draw all the text on top of the background

	currentPosition=beginningPosition;				// restore these from above
	theChunk=beginningChunk;
	theOffset=beginningOffset;
	gceColumnNumber=beginningColumn;

	gceBufferIndex=0;								// begin at start of buffer
	gceRunningWidth=0;

	textXOffset=xWindowOffset+distanceToStart;		// move the offset slightly back to beginning of first character which enters the view (or forward if view is scrolled negatively)

	currentSelection=GetSelectionAtPosition(theView->parentBuffer->selectionUniverse,currentPosition);
	currentStyle=GetStyleAtPosition(theView->parentBuffer->styleUniverse,currentPosition);
	ForceStyleIntoRange(currentStyle);
	gceWidthTable=&(GetFontForStyle(theView->viewStyles,currentStyle)->charWidths[0]);

	while(theChunk&&((textXOffset+gceRunningWidth)<xWindowEndOffset)&&GetCharacterEquiv(theChunk->data[theOffset]))		// fill in all of the text needed
	{
		nextSelection=GetSelectionAtPosition(theView->parentBuffer->selectionUniverse,currentPosition+1);
		nextStyle=GetStyleAtPosition(theView->parentBuffer->styleUniverse,currentPosition+1);
		ForceStyleIntoRange(nextStyle);

		if((currentStyle!=nextStyle)||(currentSelection!=nextSelection)||(gceBufferIndex>STAGING_BUFFER_SIZE-STAGING_MAX_ADD))	// is something about to change? (if so, output what we have so far) (this will also output if the staging buffer is getting full)
		{
			DrawViewTextPiece(theView,currentSelection,currentStyle,textXOffset,gceRunningWidth,yWindowOffset,xWindowEndOffset);

			textXOffset+=gceRunningWidth;

			gceBufferIndex=0;						// begin at start of buffer
			gceRunningWidth=0;

			currentSelection=nextSelection;
			currentStyle=nextStyle;
			gceWidthTable=&(GetFontForStyle(theView->viewStyles,currentStyle)->charWidths[0]);
		}

		currentPosition++;
		theOffset++;
		if(theOffset>=theChunk->totalBytes)			// push through end of chunk
		{
			theChunk=theChunk->nextHeader;
			theOffset=0;
		}
	}
	// ran off the end of the line, or the end of the universe
	if(gceBufferIndex)
	{
		DrawViewTextPiece(theView,currentSelection,currentStyle,textXOffset,gceRunningWidth,yWindowOffset,xWindowEndOffset);
	}
}

static void DrawViewLines(EDITOR_VIEW *theView)
// Draw the lines into theView, obey the invalid region of the view
{
	GUI_VIEW_INFO
		*theViewInfo;
	XRectangle
		lineRect,
		regionBox;
	Region
		invalidTextRegion,
		lineRegion;
	UINT32
		topLineToDraw,
		bottomLineToDraw;
	INT32
		xWindowOffset;
	INT32
		xWindowEndOffset;
	CHUNK_HEADER
		*theChunk;
	UINT32
		theOffset,
		currentPosition;
	UINT32
		i;

	theViewInfo=theView->viewInfo;			// point at the information record for this view

	gceTabSize=theView->tabSize;

	regionBox.x=lineRect.x=theViewInfo->textBounds.x;
	regionBox.width=lineRect.width=theViewInfo->textBounds.w;
	lineRect.height=theViewInfo->lineHeight;
	regionBox.y=theViewInfo->textBounds.y;
	regionBox.height=theViewInfo->textBounds.h;

	invalidTextRegion=XCreateRegion();
	XUnionRectWithRegion(&regionBox,invalidTextRegion,invalidTextRegion);		// make rectangular region of entire text space
	XIntersectRegion(invalidTextRegion,((WINDOW_LIST_ELEMENT *)theView->parentWindow->userData)->invalidRegion,invalidTextRegion);	// intersect with invalid region of window

	XClipBox(invalidTextRegion,&regionBox);				// get bounds for actual invalid text area

	topLineToDraw=(regionBox.y-theViewInfo->textBounds.y)/theViewInfo->lineHeight;	// screen relative top line to draw
	bottomLineToDraw=(regionBox.y-theViewInfo->textBounds.y+regionBox.height-1)/theViewInfo->lineHeight;	// get last line to draw

	for(i=topLineToDraw;i<=bottomLineToDraw;i++)
	{
		lineRect.y=theViewInfo->textBounds.y+i*theViewInfo->lineHeight;
		lineRegion=XCreateRegion();
		XUnionRectWithRegion(&lineRect,lineRegion,lineRegion);		// make region for the line we are about to draw
		XIntersectRegion(lineRegion,invalidTextRegion,lineRegion);	// see if it intersects with the invalid region
		if(!XEmptyRegion(lineRegion))
		{
			XClipBox(lineRegion,&regionBox);						// get bounds for this part
			LineToChunkPosition(theView->parentBuffer->textUniverse,theViewInfo->topLine+i,&theChunk,&theOffset,&currentPosition);	// point to data for this line

			xWindowOffset=regionBox.x;
			xWindowEndOffset=regionBox.x+regionBox.width;

			// extend the draw area slightly to account for left and right text overhang

			xWindowOffset-=theViewInfo->maxRightOverhang;			// move back far enough to let any character which extends to the right to draw what it needs
			xWindowEndOffset+=theViewInfo->maxLeftOverhang;			// do same for right hand side

			DrawViewLine(theView,currentPosition,xWindowOffset,xWindowEndOffset,lineRect.y);
		}
		XDestroyRegion(lineRegion);
	}
	XDestroyRegion(invalidTextRegion);
}

void DrawView(EDITOR_VIEW *theView)
// Draw the given view in its parent window
// also handle highlighting of text
// NOTE: the invalidRegion of the view's parent
// window should be set to the region that needs update
{
	XRectangle
		regionBox;
	GUI_VIEW_INFO
		*theViewInfo;
	WINDOW_LIST_ELEMENT
		*theWindowElement;
	GC
		graphicsContext;
	Region
		viewInvalidRegion;

	theViewInfo=theView->viewInfo;										// point at the information record for this view

	theWindowElement=(WINDOW_LIST_ELEMENT *)theView->parentWindow->userData;	// get the x window information associated with this editor window
	graphicsContext=theWindowElement->graphicsContext;					// get graphics context for this window

	viewInvalidRegion=XCreateRegion();
	regionBox.x=theViewInfo->bounds.x;
	regionBox.y=theViewInfo->bounds.y;
	regionBox.width=theViewInfo->bounds.w;
	regionBox.height=theViewInfo->bounds.h;
	XUnionRectWithRegion(&regionBox,viewInvalidRegion,viewInvalidRegion);	// create a region which is the rectangle of the entire view
	XIntersectRegion(viewInvalidRegion,theWindowElement->invalidRegion,viewInvalidRegion);	// intersect with what is invalid in this window

	if(!XEmptyRegion(viewInvalidRegion))								// see if it has an invalid area
	{
		XClipBox(viewInvalidRegion,&regionBox);							// get bounds of invalid region

		XSetRegion(xDisplay,graphicsContext,viewInvalidRegion);			// set clipping to what's invalid in the view

		XSetFillStyle(xDisplay,graphicsContext,FillSolid);
		XSetForeground(xDisplay,graphicsContext,EditorColorToXPixel(theView->viewStyles[0].backgroundColor));

		if(regionBox.x<theViewInfo->textBounds.x)
		{
			XFillRectangle(xDisplay,theWindowElement->xWindow,graphicsContext,theViewInfo->bounds.x,theViewInfo->bounds.y,theViewInfo->textBounds.x-theViewInfo->bounds.x,theViewInfo->bounds.h);
		}

		if(regionBox.y<theViewInfo->textBounds.y)
		{
			XFillRectangle(xDisplay,theWindowElement->xWindow,graphicsContext,theViewInfo->bounds.x,theViewInfo->bounds.y,theViewInfo->bounds.w,theViewInfo->textBounds.y-theViewInfo->bounds.y);
		}

		if((regionBox.y+(int)regionBox.height)>(theViewInfo->textBounds.y+(int)theViewInfo->textBounds.h))
		{
			XFillRectangle(xDisplay,theWindowElement->xWindow,graphicsContext,theViewInfo->bounds.x,theViewInfo->textBounds.y+theViewInfo->textBounds.h,theViewInfo->bounds.w,(theViewInfo->bounds.y+theViewInfo->bounds.h)-(theViewInfo->textBounds.y-theViewInfo->textBounds.h));
		}

		DrawViewLines(theView);
		DrawViewCursor(theView);
		XSetClipMask(xDisplay,graphicsContext,None);					// get rid of clip mask
	}
	XDestroyRegion(viewInvalidRegion);									// get rid of invalid region for the view
}

static void RecalculateViewFontInfo(EDITOR_VIEW *theView)
// Recalculate the cached information for theView
// from its current size and set of styles
{
	INT32
		viewHeight;
	UINT32
		maxDescent;
	UINT32
		i;

	maxDescent=0;
	theView->viewInfo->lineAscent=0;
	theView->viewInfo->maxLeftOverhang=0;
	theView->viewInfo->maxRightOverhang=0;
	for(i=0;i<VIEW_MAX_STYLES;i++)
	{
		if(theView->viewStyles[i].theFont)						// see if a font is assigned for this style
		{
			if(theView->viewInfo->lineAscent<(theView->viewStyles[i].theFont->ascent))
			{
				theView->viewInfo->lineAscent=theView->viewStyles[i].theFont->ascent;
			}
			if(maxDescent<(theView->viewStyles[i].theFont->descent))
			{
				maxDescent=theView->viewStyles[i].theFont->descent;
			}
			if(theView->viewInfo->maxLeftOverhang<theView->viewStyles[i].theFont->maxLeftOverhang)
			{
				theView->viewInfo->maxLeftOverhang=theView->viewStyles[i].theFont->maxLeftOverhang;
			}
			if(theView->viewInfo->maxRightOverhang<theView->viewStyles[i].theFont->maxRightOverhang)
			{
				theView->viewInfo->maxRightOverhang=theView->viewStyles[i].theFont->maxRightOverhang;
			}
		}
	}
	theView->viewInfo->lineHeight=theView->viewInfo->lineAscent+maxDescent;

	if(theView->viewInfo->lineHeight<=0)
	{
		theView->viewInfo->lineHeight=1;								// lines are minimally one pixel tall (so we do not get an infinite number of lines on the view)
	}

	theView->viewInfo->textBounds=theView->viewInfo->bounds;			// start with copy of the view bounds
	theView->viewInfo->textBounds.y+=VIEW_TOP_MARGIN;					// leave a small margin at the top

	if(theView->viewInfo->bounds.h>VIEW_TOP_MARGIN+VIEW_BOTTOM_MARGIN)
	{
		viewHeight=theView->viewInfo->bounds.h-VIEW_TOP_MARGIN-VIEW_BOTTOM_MARGIN;	// find out how much can be used for the text
		theView->viewInfo->numLines=viewHeight/theView->viewInfo->lineHeight;	// get integer number of lines in this view
		theView->viewInfo->textBounds.h=theView->viewInfo->numLines*theView->viewInfo->lineHeight;	// alter the textBounds to be the exact rectangle to hold the text
	}
	else
	{
		theView->viewInfo->numLines=0;
		theView->viewInfo->textBounds.h=0;							// no room to draw text
	}
}

void SetViewBounds(EDITOR_VIEW *theView,EDITOR_RECT *newBounds)
// Set the bounds of theView to newBounds
// invalidate area where the view used to be but now is not
// and invalidate new areas of the view
{
	GUI_VIEW_INFO
		*theViewInfo;
	EDITOR_RECT
		*viewBounds;
	unsigned int
		theWidth,
		theHeight;

	theViewInfo=theView->viewInfo;								// point at the information record for this view

	viewBounds=&(theViewInfo->bounds);							// point at the view bounds rect

	if((newBounds->x!=viewBounds->x)||(newBounds->y!=viewBounds->y)||(newBounds->w!=viewBounds->w)||(newBounds->h!=viewBounds->h))
	{
		InvalidateWindowRect(theView->parentWindow,viewBounds);			// invalidate old area
		InvalidateWindowRect(theView->parentWindow,newBounds);			// invalidate new area

		(*viewBounds)=*newBounds;								// copy new rect into place
		RecalculateViewFontInfo(theView);						// update additional information

		if(!(theWidth=theViewInfo->textBounds.w))
		{
			theWidth=1;
		}
		if(!(theHeight=theViewInfo->textBounds.h))
		{
			theHeight=1;
		}
		XMoveResizeWindow(xDisplay,theViewInfo->cursorWindow,theViewInfo->textBounds.x,theViewInfo->textBounds.y,theWidth,theHeight);	// adjust the cursor change window
	}
}

void InvalidateViewPortion(EDITOR_VIEW *theView,UINT32 startLine,UINT32 endLine,UINT32 startPixel,UINT32 endPixel)
// Invalidate the area of the view between startLine, endLine, and startPixel, endPixel
// NOTE: startLine, endLine, startPixel, and endPixel are all relative to the displayed view
// so that startLine of 0 always means the top line, and startPixel of 0 always
// means the left-most pixel
// NOTE: startLine must be <= endLine, startPixel must be <= endPixel
// the invalidated area is exclusive of endLine, and endPixel
{
	UINT32
		totalPixels;
	GUI_VIEW_INFO
		*theViewInfo;
	EDITOR_RECT
		invalidRect;

	theViewInfo=theView->viewInfo;						// point at the information record for this view

	totalPixels=theViewInfo->textBounds.w;
	if(theViewInfo->numLines&&totalPixels)				// make sure we can actually invalidate
	{
		if(startLine>theViewInfo->numLines)
		{
			startLine=theViewInfo->numLines;
		}
		if(endLine>theViewInfo->numLines)
		{
			endLine=theViewInfo->numLines;
		}

		if(startPixel>totalPixels)
		{
			startPixel=totalPixels;
		}
		if(endPixel>totalPixels)
		{
			endPixel=totalPixels;
		}
		invalidRect.x=theViewInfo->textBounds.x+startPixel;
		invalidRect.y=theViewInfo->textBounds.y+startLine*theViewInfo->lineHeight;
		invalidRect.w=endPixel-startPixel;
		invalidRect.h=(endLine-startLine)*theViewInfo->lineHeight;
		InvalidateWindowRect(theView->parentWindow,&invalidRect);
	}
}

void InvalidateViewSelection(EDITOR_VIEW *theView,SELECTION_UNIVERSE *theSelection)
// Passed a view, and a selection, invalidate the visible area of the view
// covered by theSelection
// NOTE: this (and AddSelectionLineToRegion) should probably be moved up into
// the editor independent code. However, currently it is difficult to make it
// run as fast at that level.
{
	GUI_VIEW_INFO
		*theViewInfo;
	Region
		newSelectionRegion;
	CHUNK_HEADER
		*theChunk;
	UINT32
		theOffset,
		currentPosition;
	UINT32
		i;

	theViewInfo=(GUI_VIEW_INFO *)theView->viewInfo;			// point at the information record for this view

	newSelectionRegion=XCreateRegion();						// create a region to hold what is now in the selection

	gceTabSize=theView->tabSize;
	for(i=0;i<theViewInfo->numLines;i++)
	{
		LineToChunkPosition(theView->parentBuffer->textUniverse,theViewInfo->topLine+i,&theChunk,&theOffset,&currentPosition);	// point to data for line
		AddSelectionLineToRegion(theView,theSelection,newSelectionRegion,currentPosition,theViewInfo->textBounds.y+i*theViewInfo->lineHeight);
	}
	InvalidateWindowRegion(theView->parentWindow,newSelectionRegion);

	XDestroyRegion(newSelectionRegion);						// get rid of the region we just created
}

void InvalidateView(EDITOR_VIEW *theView)
// invalidate the entire area of the view
{
	InvalidateWindowRect(theView->parentWindow,&theView->viewInfo->bounds);
}

void ScrollViewPortion(EDITOR_VIEW *theView,UINT32 startLine,UINT32 endLine,INT32 numLines,UINT32 startPixel,UINT32 endPixel,INT32 numPixels)
// scroll the area of the view between startLine, endLine, and startPixel, endPixel
// by numLines, and numPixels
// if numLines > 0, the data is scrolled down, else it is scrolled up
// if numPixels > 0, the data is scrolled right, else it is scrolled left
// any area left open by the scroll will be left erased, and invalidated
// NOTE: startLine, endLine, startPixel, and endPixel are all relative to the displayed view
// so that startLine of 0 always means the top line, and startPixel of 0 always
// means the left-most pixel
// NOTE: startLine must be <= endLine, startPixel must be <= endPixel
// the scrolled area is exclusive of endLine, and endPixel
{
	Window
		xWindow;
	GC
		graphicsContext;
	UINT32
		totalPixels;
	INT32
		copySrcX,					// source x and y for copy
		copySrcY,
		copyDestX,
		copyDestY,
		invalidX,
		invalidY;
	UINT32
		copyWidth,
		copyHeight,
		invalidWidth,
		invalidHeight;
	EDITOR_RECT
		invalidRect;
	GUI_VIEW_INFO
		*theViewInfo;

	theViewInfo=theView->viewInfo;										// point at the information record for this view
	xWindow=((WINDOW_LIST_ELEMENT *)theView->parentWindow->userData)->xWindow;					// get x window for this view
	graphicsContext=((WINDOW_LIST_ELEMENT *)theView->parentWindow->userData)->graphicsContext;	// get graphicsContext for this view

	totalPixels=theViewInfo->textBounds.w;
	if(theViewInfo->numLines&&totalPixels)								// make sure we can actually scroll
	{
		if(startLine>theViewInfo->numLines)
		{
			startLine=theViewInfo->numLines;
		}
		if(endLine>theViewInfo->numLines)
		{
			endLine=theViewInfo->numLines;
		}

		if(startPixel>totalPixels)
		{
			startPixel=totalPixels;
		}
		if(endPixel>totalPixels)
		{
			endPixel=totalPixels;
		}

		if(numLines>((INT32)(endLine-startLine)))						// allow a scroll that erases everything
		{
			numLines=endLine-startLine;
		}
		else
		{
			if(numLines<((INT32)startLine-(INT32)endLine))
			{
				numLines=(INT32)startLine-(INT32)endLine;
			}
		}

		if(numPixels>((INT32)(endPixel-startPixel)))
		{
			numPixels=endPixel-startPixel;
		}
		else
		{
			if(numPixels<(INT32)startPixel-(INT32)endPixel)
			{
				numPixels=(INT32)startPixel-(INT32)endPixel;
			}
		}
		if(numLines||numPixels)
		{
			if(numLines>=0)												// see if scrolling down, or not scrolling
			{
				invalidY=copySrcY=startLine;
				copyDestY=startLine+numLines;
				copyHeight=(endLine-startLine)-numLines;
				invalidHeight=numLines;
			}
			else														// scrolling up
			{
				copySrcY=startLine-numLines;
				copyDestY=startLine;
				copyHeight=endLine-copySrcY;
				invalidY=endLine+numLines;
				invalidHeight=-numLines;
			}
			if(numPixels>=0)											// scrolling right, or not scrolling horizontally
			{
				invalidX=copySrcX=startPixel;
				copyDestX=startPixel+numPixels;
				copyWidth=(endPixel-startPixel)-numPixels;
				invalidWidth=numPixels;
			}
			else														// left
			{
				copySrcX=startPixel-numPixels;
				copyDestX=startPixel;
				copyWidth=endPixel-copySrcX;
				invalidX=endPixel+numPixels;
				invalidWidth=-numPixels;
			}
			if(copyWidth&&copyHeight)
			{
				XSetGraphicsExposures(xDisplay,graphicsContext,True);	// tell x to give events when part of scroll is revealed
				XGrabServer(xDisplay);									// keep anyone else from invalidating while the scroll is in progress (so they do not invalidate BEFORE the scroll, causing invalid data to be scrolled and never updated)
				XSync(xDisplay,False);									// get all invalidations
				UpdateEditorWindows();									// redraw everything which is invalid before scrolling
				XCopyArea(xDisplay,xWindow,xWindow,graphicsContext,theViewInfo->textBounds.x+copySrcX,theViewInfo->textBounds.y+copySrcY*theViewInfo->lineHeight,copyWidth,copyHeight*theViewInfo->lineHeight,theViewInfo->textBounds.x+copyDestX,theViewInfo->textBounds.y+copyDestY*theViewInfo->lineHeight);
				XUngrabServer(xDisplay);								// let go after the scroll
				XSetGraphicsExposures(xDisplay,graphicsContext,False);
			}
			if(invalidWidth)
			{
				invalidRect.x=theViewInfo->textBounds.x+invalidX;
				invalidRect.y=theViewInfo->textBounds.y+startLine*theViewInfo->lineHeight;
				invalidRect.w=invalidWidth;
				invalidRect.h=(endLine-startLine)*theViewInfo->lineHeight;
				InvalidateWindowRect(theView->parentWindow,&invalidRect);
			}
			if(invalidHeight)
			{
				invalidRect.x=theViewInfo->textBounds.x+startPixel;
				invalidRect.y=theViewInfo->textBounds.y+invalidY*theViewInfo->lineHeight;
				invalidRect.w=endPixel-startPixel;
				invalidRect.h=invalidHeight*theViewInfo->lineHeight;
				InvalidateWindowRect(theView->parentWindow,&invalidRect);
			}
		}
	}
}

void SetViewTopLeft(EDITOR_VIEW *theView,UINT32 newTopLine,INT32 newLeftPixel)
// set the upper left corner of theView
// NOTE: newTop, and newLeft will be forced into bounds by this routine
{
	UINT32
		oldTopLine;
	INT32
		oldLeftPixel;
	UINT32
		numLines;
	GUI_VIEW_INFO
		*theViewInfo;

	theViewInfo=theView->viewInfo;										// point at the information record for this view
	numLines=theView->parentBuffer->textUniverse->totalLines;
	oldTopLine=theViewInfo->topLine;
	oldLeftPixel=theViewInfo->leftPixel;

	if(newTopLine>numLines)
	{
		newTopLine=numLines;
	}
	if(newLeftPixel>EDIT_MAX_WIDTH)
	{
		newLeftPixel=EDIT_MAX_WIDTH;
	}
	if(newLeftPixel<-EDIT_MAX_WIDTH)
	{
		newLeftPixel=-EDIT_MAX_WIDTH;
	}
	ScrollViewPortion(theView,0,theViewInfo->numLines,oldTopLine-newTopLine,0,theViewInfo->textBounds.w,oldLeftPixel-newLeftPixel);
	theViewInfo->topLine=newTopLine;
	theViewInfo->leftPixel=newLeftPixel;
	if(theView->viewPositionChangedVector)								// see if someone needs to be called because of my change
	{
		theView->viewPositionChangedVector(theView);					// call him with my pointer
	}
	UpdateEditorWindows();												// redraw now to make the update more crisp
}

void SetEditorViewTopLine(EDITOR_VIEW *theView,UINT32 lineNumber)
// set the line number of the top line that is
// displayed in the given edit view
// NOTE: this does no invalidation of the view
{
	theView->viewInfo->topLine=lineNumber;
}

static void ViewEndCursorChange(EDITOR_VIEW *theView)
// when the state (not position) of the cursor
// has changed, call this, and it will update the cursor as needed
{
	GUI_VIEW_INFO
		*theViewInfo;

	theViewInfo=(GUI_VIEW_INFO *)theView->viewInfo;				// point at the information record for this view
	if(theViewInfo->oldCursorShowing!=CursorShowing(theView))	// see if the state has changed
	{
		InvertViewCursor(theView);								// if it is showing, then draw it
	}
}

static void ViewStartCursorChange(EDITOR_VIEW *theView)
// when the state (not position) of the cursor
// is about to change, call this
{
	GUI_VIEW_INFO
		*theViewInfo;

	theViewInfo=(GUI_VIEW_INFO *)theView->viewInfo;				// point at the information record for this view
	theViewInfo->oldCursorShowing=CursorShowing(theView);		// remember if it was showing
}

static void BlinkViewCursor(EDITOR_VIEW *theView)
// calling this will cause the cursor in theView to change blink states, updating it as needed
{
	GUI_VIEW_INFO
		*theViewInfo;

	theViewInfo=(GUI_VIEW_INFO *)theView->viewInfo;			// point at the information record for this view
	ViewStartCursorChange(theView);
	theViewInfo->blinkState=!theViewInfo->blinkState;		// invert blink state
	ViewEndCursorChange(theView);
}

void ResetEditorViewCursorBlink(EDITOR_VIEW *theView)
// when this is called, it will reset the cursor blinking, forcing the
// cursor to the blink-on state, and resetting the blink timeout
{
	GUI_VIEW_INFO
		*theViewInfo;

	theViewInfo=(GUI_VIEW_INFO *)theView->viewInfo;			// point at the information record for this view
	if(!theViewInfo->blinkState)							// if it is currently off, blink it to on
	{
		BlinkViewCursor(theView);
	}
	ResetCursorBlinkTime();
}

void ViewCursorTask(EDITOR_VIEW *theView)
// handle blinking the cursor for theView
{
	BlinkViewCursor(theView);
}

static bool HandleViewMouseWheelEvents(EDITOR_VIEW *theView,XEvent *theEvent)
// Added 2/12/2004 FEA
// Call this to see if buttons 4 or 5 have been clicked. X sends button press events for buttons 4 and 5 when you
// scroll the mouse wheel. Returns true if it was a wheel event, else false
{
	UINT32
		delta;
	UINT32
		i;
	XEvent
		xevent;

	if(theEvent->xbutton.button==Button4||theEvent->xbutton.button==Button5)
	{
		delta=2;
		//
		// check for other events of the same kind in the event queue and pull them off now
		// the more we have, the more we should scroll
		//
		while(XCheckTypedWindowEvent(xDisplay,((WINDOW_LIST_ELEMENT *)theView->parentWindow->userData)->xWindow,ButtonPress,&xevent))
		{
			if(xevent.xbutton.button!=theEvent->xbutton.button)
			{
				XPutBackEvent(xDisplay,&xevent);
				break;
			}
			delta++;
		}

		if(delta>1)
		{
			delta*=3;
		}
		if(theEvent->xbutton.button==Button4)
		{
			for(i=0;i<delta;i++)
			{
				ScrollEditViewEvent(theView,VVK_SCROLLUP,i);
				UpdateEditorWindows();
			}
		}
		else
		{
			for(i=0;i<delta;i++)
			{
				ScrollEditViewEvent(theView,VVK_SCROLLDOWN,i);
				UpdateEditorWindows();
			}
		}
		return(true);
	}
	return(false);
}

bool TrackView(EDITOR_VIEW *theView,XEvent *theEvent)
// call this to see if the click was in the view, and if so, track it
// if the click was not in the view, return false
{
	VIEW_EVENT
		theRecord;
	VIEW_CLICK_EVENT_DATA
		theClickData;
	GUI_VIEW_INFO
		*theViewInfo;
	Window
		root,
		child;
	int
		rootX,
		rootY,
		windowX,
		windowY;
	unsigned int
		state;
	UINT32
		numClicks,
		keyCode,
		modifiers;

	if(PointInView(theView,theEvent->xbutton.x,theEvent->xbutton.y))
	{
		if(!HandleViewMouseWheelEvents(theView,theEvent))
		{
			numClicks=HandleMultipleClicks(theEvent);
			XStateToEditorModifiers(theEvent->xbutton.state,numClicks,&modifiers);
			switch(theEvent->xbutton.button)
			{
				case Button1:
					keyCode=0;
					break;
				case Button2:
					keyCode=1;
					break;
				case Button3:
					keyCode=2;
					break;
				default:
					keyCode=0;
					break;
			}
			theClickData.keyCode=keyCode;
			theClickData.modifiers=modifiers;

			theRecord.eventType=VET_CLICK;
			theRecord.theView=theView;
			theRecord.eventData=&theClickData;
			theViewInfo=(GUI_VIEW_INFO *)(theView->viewInfo);
			LocalClickToViewClick(theRecord.theView,theEvent->xbutton.x,theEvent->xbutton.y,&theClickData.xClick,&theClickData.yClick);
			if((theClickData.xClick>=0)&&(theClickData.xClick<(int)theViewInfo->textBounds.w))	// make sure initial click is in the text part of the view
			{
				if((theClickData.yClick>=0)&&(theClickData.yClick<(int)theViewInfo->numLines))	// make sure initial click is in the view
				{
					HandleViewEvent(&theRecord);
					while(StillDown(theEvent->xbutton.button,1))							// send repeat events for the view click
					{
						if(XQueryPointer(xDisplay,((WINDOW_LIST_ELEMENT *)theView->parentWindow->userData)->xWindow,&root,&child,&rootX,&rootY,&windowX,&windowY,&state))
						{
							theRecord.theView=theView;
							theRecord.eventType=VET_CLICKHOLD;
							theRecord.eventData=&theClickData;
							theClickData.modifiers=keyCode;
							theClickData.modifiers=modifiers;
							LocalClickToViewClick(theRecord.theView,windowX,windowY,&theClickData.xClick,&theClickData.yClick);
							HandleViewEvent(&theRecord);
						}
						UpdateEditorWindows();
					}
				}
			}
		}
		return(true);
	}
	return(false);
}

void ViewStyleFontsChanged(EDITOR_VIEW *theView)
// When the fonts have been modified in styles for the passed view,
// this is called so that it may update the line height, and anything else
// that may be necessary.
{
	RecalculateViewFontInfo(theView);
}

void ViewEndChange(EDITOR_VIEW *theView)
// The passed view has finished being changed. Bring the view back
// up to date
{
	ViewStartCursorChange(theView);
	theView->viewInfo->cursorOn=true;							// turn cursor back on when done
	ViewEndCursorChange(theView);
}

void ViewStartChange(EDITOR_VIEW *theView)
// The passed view is about to be changed in some way, do whatever is needed
// so that when ViewEndChange is called, the correct parts of the view are
// invalidated.
{
	ViewStartCursorChange(theView);
	theView->viewInfo->cursorOn=false;							// turn cursor off while things are changing
	ViewEndCursorChange(theView);
}

void DisposeViewInfo(EDITOR_VIEW *theView)
// unlink theView from everywhere, and dispose of it
{
	XDestroyWindow(xDisplay,theView->viewInfo->cursorWindow);	// tell x to make the cursor window go away
	MDisposePtr(theView->viewInfo);
}

bool CreateViewInfo(EDITOR_VIEW *theView,EDITOR_VIEW_DESCRIPTOR *theDescriptor)
// Called after all editor specific fields of a new view have been filled in
// This creates and stuffs local GUI specific information, and sets up the view in the GUI
// if there is a problem, SetError, and return false
{
	unsigned int
		theWidth,
		theHeight;

	if((theView->viewInfo=(GUI_VIEW_INFO *)MNewPtr(sizeof(GUI_VIEW_INFO))))
	{
		theView->viewInfo->topLine=theDescriptor->topLine;
		theView->viewInfo->leftPixel=theDescriptor->leftPixel;		// initial pixel to display
		theView->viewInfo->bounds=theDescriptor->theRect;			// copy bounds rectangle

		theView->viewInfo->cursorOn=true;							// the cursor is turned on
		theView->viewInfo->blinkState=false;						// next time it blinks, it will blink to on
		theView->viewInfo->nextBlinkTime=0;							// next time it checks, it will blink

		RecalculateViewFontInfo(theView);							// update view information based on fonts and stuff

		if(!(theWidth=theView->viewInfo->textBounds.w))
		{
			theWidth=1;
		}
		if(!(theHeight=theView->viewInfo->textBounds.h))
		{
			theHeight=1;
		}

		// create a window which has the sole purpose of changing the cursor to a caret when it enters the view
		theView->viewInfo->cursorWindow=XCreateWindow(xDisplay,((WINDOW_LIST_ELEMENT *)theView->parentWindow->userData)->xWindow,theView->viewInfo->textBounds.x,theView->viewInfo->textBounds.y,theWidth,theHeight,0,0,InputOnly,DefaultVisual(xDisplay,xScreenNum),0,NULL);
		XDefineCursor(xDisplay,theView->viewInfo->cursorWindow,caretCursor);	// this is the cursor to use in the view
		XMapRaised(xDisplay,theView->viewInfo->cursorWindow);		// map window to the top of the pile

		return(theView);
	}
	return(false);
}

